#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017-2020 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Subsytem handling localization-related data communication
"""
import collections
import logging
import struct

from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort
from cflib.utils.callbacks import Caller
from cflib.utils.fp16 import fp16_to_float

__author__ = 'Bitcraze AB'
__all__ = ['Localization', 'LocalizationPacket']

logger = logging.getLogger(__name__)

# A generic location packet contains type and data. When received the data
# may be decoded by the lib.
LocalizationPacket = collections.namedtuple('localizationPacket',
                                            ['type', 'raw_data', 'data'])


class Localization():
    """
    Handle localization-related data communication with the Crazyflie
    """

    # Implemented channels
    POSITION_CH = 0
    GENERIC_CH = 1

    # Location message types for generig channel
    RANGE_STREAM_REPORT = 0
    RANGE_STREAM_REPORT_FP16 = 1
    LPS_SHORT_LPP_PACKET = 2
    EMERGENCY_STOP = 3
    EMERGENCY_STOP_WATCHDOG = 4
    COMM_GNSS_NMEA = 6
    COMM_GNSS_PROPRIETARY = 7
    EXT_POSE = 8
    EXT_POSE_PACKED = 9
    LH_ANGLE_STREAM = 10
    LH_PERSIST_DATA = 11

    def __init__(self, crazyflie=None):
        """
        Initialize the Extpos object.
        """
        self._cf = crazyflie

        self.receivedLocationPacket = Caller()
        self._cf.add_port_callback(CRTPPort.LOCALIZATION, self._incoming)

    def _incoming(self, packet):
        """
        Callback for data received from the copter.
        """
        if len(packet.data) < 1:
            logger.warning('Localization packet received with incorrect' +
                           'length (length is {})'.format(len(packet.data)))
            return

        pk_type = struct.unpack('<B', packet.data[:1])[0]
        data = packet.data[1:]

        # Decoding the known packet types
        # TODO: more generic decoding scheme?
        decoded_data = None
        if pk_type == self.RANGE_STREAM_REPORT:
            if len(data) % 5 != 0:
                logger.error('Wrong range stream report data lenght')
                return
            decoded_data = {}
            raw_data = data
            for i in range(int(len(data) / 5)):
                anchor_id, distance = struct.unpack('<Bf', raw_data[:5])
                decoded_data[anchor_id] = distance
                raw_data = raw_data[5:]
        elif pk_type == self.LH_PERSIST_DATA:
            decoded_data = bool(data[0])
        elif pk_type == self.LH_ANGLE_STREAM:
            decoded_data = self._decode_lh_angle(data)

        pk = LocalizationPacket(pk_type, data, decoded_data)
        self.receivedLocationPacket.call(pk)

    def _decode_lh_angle(self, data):
        decoded_data = {}

        raw_data = struct.unpack('<Bfhhhfhhh', data)

        decoded_data['basestation'] = raw_data[0]
        decoded_data['x'] = [0, 0, 0, 0]
        decoded_data['x'][0] = raw_data[1]
        decoded_data['x'][1] = raw_data[1] - fp16_to_float(raw_data[2])
        decoded_data['x'][2] = raw_data[1] - fp16_to_float(raw_data[3])
        decoded_data['x'][3] = raw_data[1] - fp16_to_float(raw_data[4])
        decoded_data['y'] = [0, 0, 0, 0]
        decoded_data['y'][0] = raw_data[5]
        decoded_data['y'][1] = raw_data[5] - fp16_to_float(raw_data[6])
        decoded_data['y'][2] = raw_data[5] - fp16_to_float(raw_data[7])
        decoded_data['y'][3] = raw_data[5] - fp16_to_float(raw_data[8])

        return decoded_data

    def send_extpos(self, pos):
        """
        Send the current Crazyflie X, Y, Z position. This is going to be
        forwarded to the Crazyflie's position estimator.
        """

        pk = CRTPPacket()
        pk.port = CRTPPort.LOCALIZATION
        pk.channel = self.POSITION_CH
        pk.data = struct.pack('<fff', pos[0], pos[1], pos[2])
        self._cf.send_packet(pk)

    def send_extpose(self, pos, quat):
        """
        Send the current Crazyflie pose (position [x, y, z] and
        attitude quaternion [qx, qy, qz, qw]). This is going to be forwarded
        to the Crazyflie's position estimator.
        """

        pk = CRTPPacket()
        pk.port = CRTPPort.LOCALIZATION
        pk.channel = self.GENERIC_CH
        pk.data = struct.pack('<Bfffffff',
                              self.EXT_POSE,
                              pos[0], pos[1], pos[2],
                              quat[0], quat[1], quat[2], quat[3])
        self._cf.send_packet(pk)

    def send_short_lpp_packet(self, dest_id, data):
        """
        Send ultra-wide-band LPP packet to dest_id
        """

        pk = CRTPPacket()
        pk.port = CRTPPort.LOCALIZATION
        pk.channel = self.GENERIC_CH
        pk.data = struct.pack('<BB', self.LPS_SHORT_LPP_PACKET, dest_id) + data
        self._cf.send_packet(pk)

    def send_emergency_stop(self):
        """
        Send emergency stop
        """

        pk = CRTPPacket()
        pk.port = CRTPPort.LOCALIZATION
        pk.channel = self.GENERIC_CH
        pk.data = struct.pack('<B', self.EMERGENCY_STOP)
        self._cf.send_packet(pk)

    def send_emergency_stop_watchdog(self):
        """
        Send emergency stop watchdog
        """

        pk = CRTPPacket()
        pk.port = CRTPPort.LOCALIZATION
        pk.channel = self.GENERIC_CH
        pk.data = struct.pack('<B', self.EMERGENCY_STOP_WATCHDOG)
        self._cf.send_packet(pk)

    def send_lh_persist_data_packet(self, geo_list, calib_list):
        """
        Send geometry and calibration data to persistent memory subsystem
        """

        geo_list.sort()
        calib_list.sort()
        max_bs_nr = 15
        if len(geo_list) > 0:
            if geo_list[0] < 0 or geo_list[-1] > max_bs_nr:
                raise Exception('Geometry BS list is not valid')
        if len(calib_list) > 0:
            if calib_list[0] < 0 or calib_list[-1] > max_bs_nr:
                raise Exception('Calibration BS list is not valid')

        mask_geo = 0
        mask_calib = 0
        for bs in geo_list:
            mask_geo += 1 << bs
        for bs in calib_list:
            mask_calib += 1 << bs

        pk = CRTPPacket()
        pk.port = CRTPPort.LOCALIZATION
        pk.channel = self.GENERIC_CH
        pk.data = struct.pack(
            '<BHH', self.LH_PERSIST_DATA, mask_geo, mask_calib)
        self._cf.send_packet(pk)

        return pk.data



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 - 2020 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
import struct
from binascii import crc32

from .memory_element import MemoryElement

logger = logging.getLogger(__name__)


class OWElement(MemoryElement):
    """Memory class with extra functionality for 1-wire memories"""

    element_mapping = {
        1: 'Board name',
        2: 'Board revision',
        3: 'Custom'
    }

    def __init__(self, id, type, size, addr, mem_handler):
        """Initialize the memory with good defaults"""
        super(OWElement, self).__init__(id=id, type=type, size=size,
                                        mem_handler=mem_handler)
        self.addr = addr

        self.valid = False

        self.vid = None
        self.pid = None
        self.name = None
        self.pins = None
        self.elements = {}

        self._update_finished_cb = None
        self._write_finished_cb = None

        self._rev_element_mapping = {}
        for key in list(OWElement.element_mapping.keys()):
            self._rev_element_mapping[OWElement.element_mapping[key]] = key

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            if addr == 0:
                if self._parse_and_check_header(data[0:8]):
                    if self._parse_and_check_elements(data[9:11]):
                        self.valid = True
                        self._update_finished_cb(self)
                        self._update_finished_cb = None
                    else:
                        # We need to fetch the elements, find out the length
                        (elem_ver, elem_len) = struct.unpack('BB', data[8:10])
                        self.mem_handler.read(self, 8, elem_len + 3)
                else:
                    # Call the update if the CRC check of the header fails,
                    # we're done here
                    if self._update_finished_cb:
                        self._update_finished_cb(self)
                        self._update_finished_cb = None
            elif addr == 0x08:
                if self._parse_and_check_elements(data):
                    self.valid = True
                if self._update_finished_cb:
                    self._update_finished_cb(self)
                    self._update_finished_cb = None

    def _parse_and_check_elements(self, data):
        """
        Parse and check the CRC and length of the elements part of the memory
        """
        crc = data[-1]
        test_crc = crc32(data[:-1]) & 0x0ff
        elem_data = data[2:-1]
        if test_crc == crc:
            while len(elem_data) > 0:
                (eid, elen) = struct.unpack('BB', elem_data[:2])
                self.elements[self.element_mapping[eid]] = \
                    elem_data[2:2 + elen].decode('ISO-8859-1')
                elem_data = elem_data[2 + elen:]
            return True
        return False

    def write_done(self, mem, addr):
        if self._write_finished_cb:
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def write_data(self, write_finished_cb):
        # First generate the header part
        header_data = struct.pack('<BIBB', 0xEB, self.pins, self.vid, self.pid)
        header_crc = crc32(header_data) & 0x0ff
        header_data += struct.pack('B', header_crc)

        # Now generate the elements part
        elem = bytearray()
        logger.debug(list(self.elements.keys()))
        for element in reversed(list(self.elements.keys())):
            elem_string = self.elements[element]
            key_encoding = self._rev_element_mapping[element]
            elem += struct.pack('BB', key_encoding, len(elem_string))
            elem += bytearray(elem_string.encode('ISO-8859-1'))

        elem_data = struct.pack('BB', 0x00, len(elem))
        elem_data += elem
        elem_crc = crc32(elem_data) & 0x0ff
        elem_data += struct.pack('B', elem_crc)

        data = header_data + elem_data

        self.mem_handler.write(self, 0x00,
                               struct.unpack('B' * len(data), data))

        self._write_finished_cb = write_finished_cb

    def erase(self, write_finished_cb):
        erase_data = bytes([0xFF] * 112)
        self.mem_handler.write(self, 0x00,
                               struct.unpack('B' * len(erase_data),
                                             erase_data))

        self._write_finished_cb = write_finished_cb

    def update(self, update_finished_cb):
        """Request an update of the memory content"""
        if not self._update_finished_cb:
            self._update_finished_cb = update_finished_cb
            self.valid = False
            logger.debug('Updating content of memory {}'.format(self.id))
            # Start reading the header
            self.mem_handler.read(self, 0, 11)

    def _parse_and_check_header(self, data):
        """Parse and check the CRC of the header part of the memory"""
        (start, self.pins, self.vid, self.pid, crc) = struct.unpack('<BIBBB',
                                                                    data)
        test_crc = crc32(data[:-1]) & 0x0ff
        if start == 0xEB and crc == test_crc:
            return True
        return False

    def __str__(self):
        """Generate debug string for memory"""
        return ('OW {} ({:02X}:{:02X}): {}'.format(
            self.addr, self.vid, self.pid, self.elements))

    def disconnect(self):
        self._update_finished_cb = None
        self._write_finished_cb = None



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from test.localization.lighthouse_test_base import LighthouseTestBase

import numpy as np

from cflib.localization.lighthouse_system_aligner import LighthouseSystemAligner
from cflib.localization.lighthouse_types import Pose


class TestLighthouseSystemAligner(LighthouseTestBase):
    def setUp(self):
        pass

    def test_that_transformation_is_found_for_single_points(self):
        # Fixture
        origin = (1.0, 0.0, 0.0)
        x_axis = [(1.0, 1.0, 0.0)]
        xy_plane = [(2.0, 1.0, 0.0)]

        # Test
        actual = LighthouseSystemAligner._find_transformation(origin, x_axis, xy_plane)

        # Assert
        self.assertVectorsAlmostEqual((0.0, 0.0, 0.0), actual.rotate_translate((1.0, 0.0, 0.0)))

        self.assertVectorsAlmostEqual((1.0, 0.0, 0.0), actual.rotate_translate((1.0, 1.0, 0.0)))
        self.assertVectorsAlmostEqual((0.0, 1.0, 0.0), actual.rotate_translate((0.0, 0.0, 0.0)))
        self.assertVectorsAlmostEqual((0.0, 0.0, 1.0), actual.rotate_translate((1.0, 0.0, 1.0)))

    def test_that_transformation_is_found_for_multiple_points(self):
        # Fixture
        origin = (1.0, 0.0, 0.0)
        x_axis = [(1.0, 1.0, 0.0), (1.0, 4.0, 0.0)]
        xy_plane = [(2.0, 1.0, 0.0), (3.0, -1.0, 0.0), (5.0, 0.0, 0.0)]

        # Test
        actual = LighthouseSystemAligner._find_transformation(origin, x_axis, xy_plane)

        # Assert
        self.assertVectorsAlmostEqual((0.0, 0.0, 0.0), actual.rotate_translate((1.0, 0.0, 0.0)))

        self.assertVectorsAlmostEqual((1.0, 0.0, 0.0), actual.rotate_translate((1.0, 1.0, 0.0)))
        self.assertVectorsAlmostEqual((0.0, 1.0, 0.0), actual.rotate_translate((0.0, 0.0, 0.0)))
        self.assertVectorsAlmostEqual((0.0, 0.0, 1.0), actual.rotate_translate((1.0, 0.0, 1.0)))

    def test_that_base_stations_are_rotated(self):
        # Fixture
        origin = (1.0, 0.0, 0.0)
        x_axis = [(1.0, 1.0, 0.0)]
        xy_plane = [(2.0, 1.0, 0.0)]

        bs_id = 7
        bs_poses = {bs_id: Pose.from_rot_vec(t_vec=(1.0, 0.0, 1.0))}

        # Test
        actual, transform = LighthouseSystemAligner.align(origin, x_axis, xy_plane, bs_poses)

        # Assert
        self.assertPosesAlmostEqual(Pose.from_rot_vec(
            R_vec=(0.0, 0.0, -np.pi / 2), t_vec=(0.0, 0.0, 1.0)), actual[bs_id])

    def test_that_solution_is_de_flipped(self):
        # Fixture
        origin = (0.0, 0.0, 0.0)
        x_axis = [(-1.0, 0.0, 0.0)]
        xy_plane = [(2.0, 1.0, 0.0)]

        bs_id = 7
        bs_poses = {bs_id: Pose.from_rot_vec(t_vec=(0.0, 0.0, 1.0))}
        expected = Pose.from_rot_vec(R_vec=(0.0, 0.0, np.pi), t_vec=(0.0, 0.0, 1.0))

        # Test
        actual, transform = LighthouseSystemAligner.align(origin, x_axis, xy_plane, bs_poses)

        # Assert
        self.assertPosesAlmostEqual(expected, actual[bs_id])

    def test_that_is_aligned_for_multiple_points_where_system_is_rotated_and_poins_are_fuzzy(self):
        # Fixture
        origin = (0.0, 0.0, 0.0)
        x_axis = [(-1.0, 0.0 + 0.01, 0.0), (-2.0, 0.0 - 0.02, 0.0)]
        xy_plane = [(2.0, 2.0, 0.0 + 0.02), (-3.0, -2.0, 0.0 + 0.01), (5.0, 0.0, 0.0 - 0.01)]

        # Note: Z of base stations must be positive (above the floor)
        bs_poses = {
            1: Pose.from_rot_vec(t_vec=(1.0, 0.0, 1.0)),
            2: Pose.from_rot_vec(t_vec=(0.0, 1.0, 1.0)),
            3: Pose.from_rot_vec(t_vec=(0.0, 0.0, 1.0)),
        }

        # Test
        actual, transform = LighthouseSystemAligner.align(origin, x_axis, xy_plane, bs_poses)

        # Assert
        self.assertVectorsAlmostEqual((-1.0, 0.0, 1.0), actual[1].translation, places=1)
        self.assertVectorsAlmostEqual((0.0, -1.0, 1.0), actual[2].translation, places=1)
        self.assertVectorsAlmostEqual((0.0, 0.0, 1.0), actual[3].translation, places=1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import logging
import re

from .crtpstack import CRTPPacket
from .exceptions import WrongUriType
from cflib.crtp.crtpdriver import CRTPDriver

prrt_installed = True
try:
    import prrt
except ImportError:
    prrt_installed = False

__author__ = 'Bitcraze AB'
__all__ = ['PrrtDriver']

logger = logging.getLogger(__name__)

MAX_PAYLOAD = 32
DEFAULT_TARGET_DELAY = 0.05  # unit: s
PRRT_LOCAL_PORT = 5000


class PrrtDriver(CRTPDriver):

    def __init__(self):
        CRTPDriver.__init__(self)
        self.prrt_socket = None
        self.uri = ''
        self.link_error_callback = None
        self.packet_log = None
        self.log_index = 0
        logger.info('Initialized PRRT driver.')

    def connect(self, uri, linkQualityCallback, linkErrorCallback):
        # check if the URI is a PRRT URI
        if not re.search('^prrt://', uri):
            raise WrongUriType('Not a prrt URI')

        # Check if it is a valid PRRT URI
        uri_match = re.search(r'^prrt://((?:[\d]{1,3})\.(?:[\d]{1,3})\.'
                              r'(?:[\d]{1,3})\.(?:[\d]{1,3})):([\d]{1,5})'
                              r'(?:/([\d]{1,6}))?$', uri)
        if not uri_match:
            raise Exception('Invalid PRRT URI')

        if not prrt_installed:
            raise Exception('PRRT is missing')

        self.uri = uri

        self.link_error_callback = linkErrorCallback

        address = uri_match.group(1)
        port = int(uri_match.group(2))
        target_delay_s = DEFAULT_TARGET_DELAY
        if uri_match.group(3):
            target_delay_s = int(uri_match.group(3)) / 1000

        self.prrt_socket = prrt.PrrtSocket(('0.0.0.0', PRRT_LOCAL_PORT),
                                           maximum_payload_size=MAX_PAYLOAD,
                                           target_delay=target_delay_s)
        self.prrt_socket.connect((address, port))

    def send_packet(self, pk):
        pk_bytes = bytearray([pk.get_header()]) + pk.data
        self.prrt_socket.send(pk_bytes)

    def receive_packet(self, wait=0):
        try:
            if wait == 0:
                pk_bytes, _ = self.prrt_socket.receive_asap()
            elif wait < 0:
                pk_bytes, _ = self.prrt_socket.receive_asap_wait()
            else:
                deadline = datetime.datetime.utcnow() + datetime.timedelta(
                    seconds=wait)
                pk_bytes, _ = self.prrt_socket.receive_asap_timedwait(deadline)
        except prrt.TimeoutException:
            return None

        if len(pk_bytes) <= 0:
            return None

        pk = CRTPPacket(pk_bytes[0], pk_bytes[1:])
        return pk

    def get_status(self):
        return 'No information available'

    def get_name(self):
        return 'prrt'

    def scan_interface(self, address):
        default_uri = 'prrt://10.8.0.208:5000'
        if prrt_installed:
            return [[default_uri, ''], ]
        return []

    def close(self):
        return



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2021 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from test.localization.lighthouse_test_base import LighthouseTestBase

import numpy as np

from cflib.localization import LighthouseBsVector
from cflib.localization.lighthouse_bs_vector import LighthouseBsVectors


class TestLighthouseBsVector(LighthouseTestBase):
    def setUp(self):
        pass

    def test_init_from_lh1_angles(self):
        # Fixture
        horiz = 0.123
        vert = -1.23

        # Test
        actual = LighthouseBsVector(horiz, vert)

        # Assert
        self.assertEqual(horiz, actual.lh_v1_horiz_angle)
        self.assertEqual(vert, actual.lh_v1_vert_angle)

    def test_conversion_to_lh2_angles_are_zero_straight_forward(self):
        # Fixture
        horiz = 0
        vert = 0

        # Test
        actual = LighthouseBsVector(horiz, vert)

        # Assert
        self.assertEqual(0.0, actual.lh_v2_angle_1)
        self.assertEqual(0.0, actual.lh_v2_angle_2)

    def test_conversion_to_lh2_angles_are_equal_with_vert_zero(self):
        # Fixture
        horiz = 1.0
        vert = 0.0

        # Test
        actual = LighthouseBsVector(horiz, vert)

        # Assert
        self.assertEqual(actual.lh_v2_angle_1, actual.lh_v2_angle_2)

    def test_conversion_to_from_lh2(self):
        # Fixture
        horiz = 0.123
        vert = -0.987
        v1 = LighthouseBsVector(horiz, vert)

        # Test
        actual = LighthouseBsVector.from_lh2(v1.lh_v2_angle_1, v1.lh_v2_angle_2)

        # Assert
        self.assertAlmostEqual(horiz, actual.lh_v1_horiz_angle)
        self.assertAlmostEqual(vert, actual.lh_v1_vert_angle)

    def test_conversion_to_cartesian_straight_forward(self):
        # Fixture
        horiz = 0.0
        vert = 0.0
        vector = LighthouseBsVector(horiz, vert)

        # Test
        actual = vector.cart

        # Assert
        self.assertAlmostEqual(1.0, actual[0])
        self.assertAlmostEqual(0.0, actual[1])
        self.assertAlmostEqual(0.0, actual[2])

    def test_conversion_to_from_cartesian(self):
        # Fixture
        horiz = 0.123
        vert = -0.987
        v1 = LighthouseBsVector(horiz, vert)

        # Test
        actual = LighthouseBsVector.from_cart(v1.cart)

        # Assert
        self.assertAlmostEqual(horiz, actual.lh_v1_horiz_angle)
        self.assertAlmostEqual(vert, actual.lh_v1_vert_angle)

    def test_cartesian_is_normalized(self):
        # Fixture
        horiz = 0.123
        vert = 0.456
        vector = LighthouseBsVector(horiz, vert)

        # Test
        actual = np.linalg.norm(vector.cart)

        # Assert
        self.assertAlmostEqual(1.0, actual)

    def test_conversion_to_from_projection(self):
        # Fixture
        horiz = 0.123
        vert = 0.456
        v1 = LighthouseBsVector(horiz, vert)

        # Test
        actual = LighthouseBsVector.from_projection(v1.projection)

        # Assert
        self.assertAlmostEqual(horiz, actual.lh_v1_horiz_angle)
        self.assertAlmostEqual(vert, actual.lh_v1_vert_angle)

    def test_conversion_to_projection_pair_list(self):
        # Fixture
        vectors = LighthouseBsVectors((
            LighthouseBsVector(0.0, 0.0),
            LighthouseBsVector(0.1, 0.1),
            LighthouseBsVector(0.2, 0.2),
            LighthouseBsVector(0.3, 0.3),
        ))

        # Test
        actual = vectors.projection_pair_list()

        # Assert
        self.assertEqual(len(vectors), len(actual))
        self.assertListEqual(vectors[0].projection.tolist(), actual[0].tolist())
        self.assertListEqual(vectors[3].projection.tolist(), actual[3].tolist())

    def test_conversion_to_angle_list(self):
        # Fixture
        vectors = LighthouseBsVectors((
            LighthouseBsVector(0.0, 0.1),
            LighthouseBsVector(0.2, 0.3),
            LighthouseBsVector(0.4, 0.5),
            LighthouseBsVector(0.6, 0.7),
        ))

        # Test
        actual = vectors.angle_list()

        # Assert
        self.assertVectorsAlmostEqual((0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7), actual)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from cflib.localization.lighthouse_types import LhCfPoseSample
from cflib.localization.lighthouse_types import LhMeasurement


class LighthouseSampleMatcher:
    """Utility class to match samples of measurements from multiple lighthouse base stations.

    Assuming that the Crazyflie was moving when the measurements were recorded,
    samples that were meassured aproximately at the same position are aggregated into
    a list of LhCfPoseSample. Matching is done using the timestamp and a maximum time span.
    """

    @classmethod
    def match(cls, samples: list[LhMeasurement], max_time_diff: float = 0.020,
              min_nr_of_bs_in_match: int = 0) -> list[LhCfPoseSample]:
        """
        Aggregate samples close in time into lists
        """

        result = []
        current: LhCfPoseSample = None

        for sample in samples:
            ts = sample.timestamp

            if current is None:
                current = LhCfPoseSample(timestamp=ts)

            if ts > (current.timestamp + max_time_diff):
                cls._append_result(current, result, min_nr_of_bs_in_match)
                current = LhCfPoseSample(timestamp=ts)

            current.angles_calibrated[sample.base_station_id] = sample.angles

        cls._append_result(current, result, min_nr_of_bs_in_match)
        return result

    @classmethod
    def _append_result(cls, current: LhCfPoseSample, result: list[LhCfPoseSample], min_nr_of_bs_in_match: int):
        if current is not None and len(current.angles_calibrated) >= min_nr_of_bs_in_match:
            result.append(current)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import unittest

from cflib.crtp.crtpstack import CRTPPacket


class CRTPPacketTest(unittest.TestCase):

    def setUp(self):
        self.callback_count = 0
        self.sut = CRTPPacket()

    def test_that_port_and_channle_is_encoded_in_header(self):
        # Fixture
        self.sut.set_header(2, 1)

        # Test
        actual = self.sut.get_header()

        # Assert
        expected = 0x2d
        self.assertEqual(expected, actual)

    def test_that_port_is_truncated_in_header(self):
        # Fixture
        port = 0xff
        self.sut.set_header(port, 0)

        # Test
        actual = self.sut.get_header()

        # Assert
        expected = 0xfc
        self.assertEqual(expected, actual)

    def test_that_channel_is_truncated_in_header(self):
        # Fixture
        channel = 0xff
        self.sut.set_header(0, channel)

        # Test
        actual = self.sut.get_header()

        # Assert
        expected = 0x0f
        self.assertEqual(expected, actual)

    def test_that_port_and_channel_is_encoded_in_header_when_set_separat(self):
        # Fixture
        self.sut.port = 2
        self.sut.channel = 1

        # Test
        actual = self.sut.get_header()

        # Assert
        expected = 0x2d
        self.assertEqual(expected, actual)

    def test_that_default_header_is_set_when_constructed(self):
        # Fixture

        # Test
        actual = self.sut.get_header()

        # Assert
        expected = 0x0c
        self.assertEqual(expected, actual)

    def test_that_header_is_set_when_constructed(self):
        # Fixture
        sut = CRTPPacket(header=0x21)

        # Test
        actual = sut.get_header()

        # Assert
        self.assertEqual(0x2d, actual)
        self.assertEqual(2, sut.port)
        self.assertEqual(1, sut.channel)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 - 2020 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
import struct

from .memory_element import MemoryElement

logger = logging.getLogger(__name__)


class LighthouseBsGeometry:
    """Container for geometry data of one Lighthouse base station"""

    SIZE_FLOAT = 4
    SIZE_BOOL = 1
    SIZE_VECTOR = 3 * SIZE_FLOAT
    SIZE_GEOMETRY = (1 + 3) * SIZE_VECTOR + SIZE_BOOL

    FILE_ID_ORIGIN = 'origin'
    FILE_ID_ROTATION = 'rotation'

    yaml_tag = 'LighthouseBsGeometry'

    def __init__(self):
        self.origin = [0.0, 0.0, 0.0]
        self.rotation_matrix = [
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
        ]
        self.valid = False

    def set_from_mem_data(self, data):
        self.origin = self._read_vector(
            data[0 * self.SIZE_VECTOR:1 * self.SIZE_VECTOR])
        self.rotation_matrix = [
            self._read_vector(data[1 * self.SIZE_VECTOR:2 * self.SIZE_VECTOR]),
            self._read_vector(data[2 * self.SIZE_VECTOR:3 * self.SIZE_VECTOR]),
            self._read_vector(data[3 * self.SIZE_VECTOR:4 * self.SIZE_VECTOR]),
        ]
        self.valid = struct.unpack('<?', data[4 * self.SIZE_VECTOR:])[0]

    def add_mem_data(self, data):
        self._add_vector(data, self.origin)
        self._add_vector(data, self.rotation_matrix[0])
        self._add_vector(data, self.rotation_matrix[1])
        self._add_vector(data, self.rotation_matrix[2])
        data += struct.pack('<?', self.valid)

    def _add_vector(self, data, vector):
        data += struct.pack('<fff', vector[0], vector[1], vector[2])

    def _read_vector(self, data):
        x, y, z = struct.unpack('<fff', data)
        return [x, y, z]

    def as_file_object(self):
        return {
            self.FILE_ID_ORIGIN: self.origin,
            self.FILE_ID_ROTATION: self.rotation_matrix
        }

    @classmethod
    def from_file_object(cls, file_object):
        result = cls()
        result.origin = file_object[cls.FILE_ID_ORIGIN]
        result.rotation_matrix = file_object[cls.FILE_ID_ROTATION]
        result.valid = True
        return result

    def dump(self):
        print('origin:', self.origin)
        print('rotation matrix:', self.rotation_matrix)
        print('valid:', self.valid)


class LighthouseCalibrationSweep:
    FILE_ID_PHASE = 'phase'
    FILE_ID_TILT = 'tilt'
    FILE_ID_CURVE = 'curve'
    FILE_ID_GIBMAG = 'gibmag'
    FILE_ID_GIBPHASE = 'gibphase'
    FILE_ID_OGEEMAG = 'ogeemag'
    FILE_ID_OGEEPHASE = 'ogeephase'

    def __init__(self):
        self.phase = 0.0
        self.tilt = 0.0
        self.curve = 0.0
        self.gibmag = 0.0
        self.gibphase = 0.0
        self.ogeemag = 0.0
        self.ogeephase = 0.0

    def as_file_object(self):
        return {
            self.FILE_ID_PHASE: self.phase,
            self.FILE_ID_TILT: self.tilt,
            self.FILE_ID_CURVE: self.curve,
            self.FILE_ID_GIBMAG: self.gibmag,
            self.FILE_ID_GIBPHASE: self.gibphase,
            self.FILE_ID_OGEEMAG: self.ogeemag,
            self.FILE_ID_OGEEPHASE: self.ogeephase,
        }

    @classmethod
    def from_file_object(cls, file_object):
        result = cls()

        result.phase = file_object[cls.FILE_ID_PHASE]
        result.tilt = file_object[cls.FILE_ID_TILT]
        result.curve = file_object[cls.FILE_ID_CURVE]
        result.gibmag = file_object[cls.FILE_ID_GIBMAG]
        result.gibphase = file_object[cls.FILE_ID_GIBPHASE]
        result.ogeemag = file_object[cls.FILE_ID_OGEEMAG]
        result.ogeephase = file_object[cls.FILE_ID_OGEEPHASE]

        return result

    def dump(self):
        print(('phase: {}, tilt: {}, curve: {}, gibmag: {}, ' +
               'gibphase: {}, ogeemag: {}, ogeephase: {}').format(
            self.phase,
            self.tilt,
            self.curve,
            self.gibmag,
            self.gibphase,
            self.ogeemag,
            self.ogeephase))


class LighthouseBsCalibration:
    """Container for calibration data of one Lighthouse base station"""

    SIZE_FLOAT = 4
    SIZE_UINT_32 = 4
    SIZE_BOOL = 1
    SIZE_SWEEP = 7 * SIZE_FLOAT
    SIZE_CALIBRATION = 2 * SIZE_SWEEP + SIZE_UINT_32 + SIZE_BOOL

    FILE_ID_SWEEPS = 'sweeps'
    FILE_ID_UID = 'uid'

    def __init__(self):
        self.sweeps = [LighthouseCalibrationSweep(),
                       LighthouseCalibrationSweep()]
        self.uid = 0
        self.valid = False

    def set_from_mem_data(self, data):
        self.sweeps[0] = self._unpack_sweep_calibration(
            data[0:self.SIZE_SWEEP])
        self.sweeps[1] = self._unpack_sweep_calibration(
            data[self.SIZE_SWEEP:self.SIZE_SWEEP * 2])
        self.uid, self.valid = struct.unpack('<L?', data[self.SIZE_SWEEP * 2:])

    def _unpack_sweep_calibration(self, data):
        result = LighthouseCalibrationSweep()

        (result.phase,
         result.tilt,
         result.curve,
         result.gibmag,
         result.gibphase,
         result.ogeemag,
         result.ogeephase) = struct.unpack('<fffffff', data)

        return result

    def add_mem_data(self, data):
        self._pack_sweep_calib(data, self.sweeps[0])
        self._pack_sweep_calib(data, self.sweeps[1])
        data += struct.pack('<L?', self.uid, self.valid)

    def _pack_sweep_calib(self, data, sweep_calib):
        data += struct.pack('<fffffff',
                            sweep_calib.phase,
                            sweep_calib.tilt,
                            sweep_calib.curve,
                            sweep_calib.gibmag,
                            sweep_calib.gibphase,
                            sweep_calib.ogeemag,
                            sweep_calib.ogeephase)

    def as_file_object(self):
        return {
            self.FILE_ID_SWEEPS: [self.sweeps[0].as_file_object(), self.sweeps[1].as_file_object()],
            self.FILE_ID_UID: self.uid
        }

    @classmethod
    def from_file_object(cls, file_object):
        result = cls()

        sweeps = file_object[cls.FILE_ID_SWEEPS]
        result.sweeps[0] = LighthouseCalibrationSweep.from_file_object(sweeps[0])
        result.sweeps[1] = LighthouseCalibrationSweep.from_file_object(sweeps[1])
        result.uid = file_object[cls.FILE_ID_UID]
        result.valid = True

        return result

    def dump(self):
        self.sweeps[0].dump()
        self.sweeps[1].dump()
        print('uid: {:08X}'.format(self.uid))
        print('valid: {}'.format(self.valid))


class LighthouseMemory(MemoryElement):
    """
    Memory interface for lighthouse configuration data
    """
    GEO_START_ADDR = 0x00
    CALIB_START_ADDR = 0x1000
    PAGE_SIZE = 0x100
    NUMBER_OF_BASESTATIONS = 2
    SIZE_GEOMETRY_ALL = NUMBER_OF_BASESTATIONS * \
        LighthouseBsGeometry.SIZE_GEOMETRY

    def __init__(self, id, type, size, mem_handler):
        """Initialize Lighthouse memory"""
        super(LighthouseMemory, self).__init__(id=id, type=type, size=size,
                                               mem_handler=mem_handler)

        self._clear_update_cb()
        self._clear_write_cb()

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            tmp_update_finished_cb = self._update_finished_cb
            self._clear_update_cb()

            if addr < self.CALIB_START_ADDR:
                geo_data = LighthouseBsGeometry()
                geo_data.set_from_mem_data(data)

                if tmp_update_finished_cb:
                    tmp_update_finished_cb(self, geo_data)
            else:
                calibration_data = LighthouseBsCalibration()
                calibration_data.set_from_mem_data(data)

                if tmp_update_finished_cb:
                    tmp_update_finished_cb(self, calibration_data)

    def new_data_failed(self, mem, addr, data):
        """Callback when a read failed"""
        if mem.id == self.id:
            tmp_update_failed_cb = self._update_failed_cb
            self._clear_update_cb()

            if tmp_update_failed_cb:
                logger.debug('Update of data failed')
                tmp_update_failed_cb(self)

    def read_geo_data(self, bs_id, update_finished_cb, update_failed_cb=None):
        """Request a read of geometry data for one base station"""
        if self._update_finished_cb:
            raise Exception('Read operation already ongoing')
        self._update_finished_cb = update_finished_cb
        self._update_failed_cb = update_failed_cb
        self.mem_handler.read(self, self.GEO_START_ADDR + bs_id *
                              self.PAGE_SIZE,
                              LighthouseBsGeometry.SIZE_GEOMETRY)

    def read_calib_data(self, bs_id, update_finished_cb,
                        update_failed_cb=None):
        """Request a read of calibration data for one base station"""
        if self._update_finished_cb:
            raise Exception('Read operation already ongoing')
        self._update_finished_cb = update_finished_cb
        self._update_failed_cb = update_failed_cb
        self.mem_handler.read(self, self.CALIB_START_ADDR + bs_id *
                              self.PAGE_SIZE,
                              LighthouseBsCalibration.SIZE_CALIBRATION)

    def write_geo_data(self, bs_id, geo_data, write_finished_cb,
                       write_failed_cb=None):
        """Write geometry data for one base station to the Crazyflie"""
        if self._write_finished_cb:
            raise Exception('Write operation already ongoing.')
        data = bytearray()
        geo_data.add_mem_data(data)
        self._write_finished_cb = write_finished_cb
        self._write_failed_cb = write_failed_cb
        geo_addr = self.GEO_START_ADDR + bs_id * self.PAGE_SIZE
        self.mem_handler.write(self, geo_addr, data, flush_queue=True)

    def write_calib_data(self, bs_id, calibration_data, write_finished_cb,
                         write_failed_cb=None):
        """Write calibration data for one basestation to the Crazyflie"""
        if self._write_finished_cb:
            raise Exception('Write operation already ongoing.')
        data = bytearray()
        calibration_data.add_mem_data(data)
        self._write_finished_cb = write_finished_cb
        self._write_failed_cb = write_failed_cb
        calib_addr = self.CALIB_START_ADDR + bs_id * self.PAGE_SIZE
        self.mem_handler.write(self, calib_addr, data, flush_queue=True)

    def _write_data_list(self, addr, data_list):
        data = bytearray()
        for bs in data_list:
            bs.add_mem_data(data)
        self.mem_handler.write(self, addr, data, flush_queue=True)

    def write_done(self, mem, addr):
        if mem.id == self.id:
            tmp_cb = self._write_finished_cb
            self._clear_write_cb()
            if tmp_cb:
                tmp_cb(self, addr)

    def write_failed(self, mem, addr):
        if mem.id == self.id:
            tmp_cb = self._write_failed_cb
            self._clear_write_cb()
            if tmp_cb:
                logger.debug('Write of data failed')
                tmp_cb(self, addr)

    def disconnect(self):
        self._clear_update_cb()
        self._clear_write_cb()

    def _clear_update_cb(self):
        self._update_finished_cb = None
        self._update_failed_cb = None

    def _clear_write_cb(self):
        self._write_finished_cb = None
        self._write_failed_cb = None


class LighthouseMemHelper:
    """Helper to access all geometry and calibration data located in crazyflie memory subsystem"""

    NR_OF_CHANNELS = 16

    class _ObjectReader:
        """Internal class that reads all geos or calib objects"""
        NR_OF_CHANNELS = 16

        def __init__(self, read_fcn):
            self._read_fcn = read_fcn

            self._result = None
            self._next_id = None
            self._read_done_cb = None

        def read_all(self, read_done_cb):
            if self._read_done_cb is not None:
                raise Exception('Read operation not finished')

            self._result = {}
            self._next_id = 0
            self._read_done_cb = read_done_cb
            self._get_object(0)

        def _data_updated(self, mem, data):
            self._result[self._next_id] = data
            self._next_id += 1
            self._get_object(self._next_id)

        def _update_failed(self, mem):
            # Update failes if the object is not available, that is if we try to read a base station id that is
            # not supported by the firmware. Try to read the next one until we're done.
            self._next_id += 1
            self._get_object(self._next_id)

        def _get_object(self, channel):
            if channel < self.NR_OF_CHANNELS:
                self._read_fcn(channel, self._data_updated, update_failed_cb=self._update_failed)
            else:
                tmp_cb = self._read_done_cb
                tmp_result = self._result

                self._read_done_cb = None
                self._result = None
                self._next_id = None

                tmp_cb(tmp_result)

    class _ObjectWriter:
        """Internal class that writes all geos or calib objects"""

        def __init__(self, write_fcn):
            self._objects_to_write = None
            self._write_done_cb = None
            self._write_failed_for_one_or_more_objects = False
            self._write_fcn = write_fcn

        def write(self, object_dict, write_done_cb):
            if self._objects_to_write is not None:
                raise Exception('Write operation not finished')

            self._write_done_cb = write_done_cb
            # Make a copy of the dictionary
            self._objects_to_write = dict(object_dict)
            self._write_failed_for_one_or_more_objects = False
            self._write_next_object()

        def _write_next_object(self):
            if len(self._objects_to_write) > 0:
                id = list(self._objects_to_write.keys())[0]
                data = self._objects_to_write.pop(id)
                self._write_fcn(id, data, self._data_written, write_failed_cb=self._write_failed)
            else:
                tmp_cb = self._write_done_cb
                is_sucess = not self._write_failed_for_one_or_more_objects

                self._objects_to_write = None
                self._write_done_cb = None
                self._write_failed_for_one_or_more_objects = False

                tmp_cb(is_sucess)

        def _data_written(self, mem, addr):
            self._write_next_object()

        def _write_failed(self, mem, addr):
            # Write failes if we try to write data for a base station that is not supported by the fw.
            # Try to write the next one until we have tried them all, but record the problem and
            # report that not all base stations were written.
            self._write_failed_for_one_or_more_objects = True
            self._write_next_object()

    def __init__(self, cf):
        mems = cf.mem.get_mems(MemoryElement.TYPE_LH)
        count = len(mems)
        if count != 1:
            raise Exception('Unexpected nr of memories found:', count)

        lh_mem = mems[0]

        self.geo_reader = self._ObjectReader(lh_mem.read_geo_data)
        self.geo_writer = self._ObjectWriter(lh_mem.write_geo_data)

        self.calib_reader = self._ObjectReader(lh_mem.read_calib_data)
        self.calib_writer = self._ObjectWriter(lh_mem.write_calib_data)

    def read_all_geos(self, read_done_cb):
        """
        Read geometry data for all base stations. The result is returned
        as a dictionary keyed on base station channel (0-indexed) with
        geometry data as values
        """
        self.geo_reader.read_all(read_done_cb)

    def write_geos(self, geometry_dict, write_done_cb):
        """
        Write geometry data for one or more base stations. Input is
        a dictionary keyed on base station channel (0-indexed) with
        geometry data as values. The callback is called with a boolean
        indicating if all items were successfully written.
        """
        self.geo_writer.write(geometry_dict, write_done_cb)

    def read_all_calibs(self, read_done_cb):
        """
        Read calibration data for all base stations. The result is returned
        as a dictionary keyed on base station channel (0-indexed) with
        calibration data as values
        """
        self.calib_reader.read_all(read_done_cb)

    def write_calibs(self, calibration_dict, write_done_cb):
        """
        Write calibration data for one or more base stations. Input is
        a dictionary keyed on base station channel (0-indexed) with
        calibration data as values. The callback is called with a boolean
        indicating if all items were successfully written.
        """
        self.calib_writer.write(calibration_dict, write_done_cb)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 - 2020 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging

from .memory_element import MemoryElement

logger = logging.getLogger(__name__)


class LED:
    """Used to set color/intensity of one LED in the LED-ring"""

    def __init__(self):
        """Initialize to off"""
        self.r = 0
        self.g = 0
        self.b = 0
        self.intensity = 100

    def set(self, r, g, b, intensity=None):
        """Set the R/G/B and optionally intensity in one call"""
        self.r = r
        self.g = g
        self.b = b
        if intensity:
            self.intensity = intensity


class LEDDriverMemory(MemoryElement):
    """Memory interface for using the LED-ring mapped memory for setting RGB
       values for all the LEDs in the ring"""

    def __init__(self, id, type, size, mem_handler):
        """Initialize with 12 LEDs"""
        super(LEDDriverMemory, self).__init__(id=id, type=type, size=size,
                                              mem_handler=mem_handler)
        self._update_finished_cb = None
        self._write_finished_cb = None

        self.leds = []
        for i in range(12):
            self.leds.append(LED())

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            logger.debug(
                "Got new data from the LED driver, but we don't care.")

    def write_data(self, write_finished_cb):
        """Write the saved LED-ring data to the Crazyflie"""
        self._write_finished_cb = write_finished_cb
        data = bytearray()
        for led in self.leds:
            # In order to fit all the LEDs in one radio packet RGB565 is used
            # to compress the colors. The calculations below converts 3 bytes
            # RGB into 2 bytes RGB565. Then shifts the value of each color to
            # LSB, applies the intensity and shifts them back for correct
            # alignment on 2 bytes.
            R5 = ((int)((((int(led.r) & 0xFF) * 249 + 1014) >> 11) & 0x1F) *
                  led.intensity / 100)
            G6 = ((int)((((int(led.g) & 0xFF) * 253 + 505) >> 10) & 0x3F) *
                  led.intensity / 100)
            B5 = ((int)((((int(led.b) & 0xFF) * 249 + 1014) >> 11) & 0x1F) *
                  led.intensity / 100)
            tmp = (int(R5) << 11) | (int(G6) << 5) | (int(B5) << 0)
            data += bytearray((tmp >> 8, tmp & 0xFF))
        self.mem_handler.write(self, 0x00, data, flush_queue=True)

    def update(self, update_finished_cb):
        """Request an update of the memory content"""
        if not self._update_finished_cb:
            self._update_finished_cb = update_finished_cb
            self.valid = False
            logger.debug('Updating content of memory {}'.format(self.id))
            # Start reading the header
            self.mem_handler.read(self, 0, 16)

    def write_done(self, mem, addr):
        if self._write_finished_cb and mem.id == self.id:
            logger.debug('Write to LED driver done')
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def disconnect(self):
        self._update_finished_cb = None
        self._write_finished_cb = None



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2018 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import math
import unittest
from unittest.mock import call
from unittest.mock import MagicMock
from unittest.mock import patch

from cflib.crazyflie import Crazyflie
from cflib.crazyflie import HighLevelCommander
from cflib.crazyflie import Param
from cflib.positioning.position_hl_commander import PositionHlCommander


@patch('time.sleep')
class TestPositionHlCommander(unittest.TestCase):
    def setUp(self):
        self.commander_mock = MagicMock(spec=HighLevelCommander)
        self.param_mock = MagicMock(spec=Param)
        self.cf_mock = MagicMock(spec=Crazyflie)
        self.cf_mock.high_level_commander = self.commander_mock
        self.cf_mock.param = self.param_mock
        self.cf_mock.is_connected.return_value = True

        self.sut = PositionHlCommander(self.cf_mock)

    def test_that_the_hi_level_commander_is_activated_on_creation(
            self, sleep_mock):
        # Fixture

        # Test

        # Assert
        self.param_mock.set_value.assert_has_calls([
            call('commander.enHighLevel', '1')
        ])

    def test_that_controller_is_selected_on_creation(
            self, sleep_mock):
        # Fixture

        # Test
        PositionHlCommander(
            self.cf_mock,
            controller=PositionHlCommander.CONTROLLER_MELLINGER)

        # Assert
        self.param_mock.set_value.assert_has_calls([
            call('stabilizer.controller', '2')
        ])

    def test_that_take_off_raises_exception_if_not_connected(
            self, sleep_mock):
        # Fixture
        self.cf_mock.is_connected.return_value = False

        # Test
        # Assert
        with self.assertRaises(Exception):
            self.sut.take_off()

    def test_that_take_off_raises_exception_when_already_flying(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()

        # Test
        # Assert
        with self.assertRaises(Exception):
            self.sut.take_off()

    def test_that_it_goes_up_on_take_off(
            self, sleep_mock):
        # Fixture

        # Test
        self.sut.take_off(height=0.4, velocity=0.6)

        # Assert
        duration = 0.4 / 0.6
        self.commander_mock.takeoff.assert_called_with(0.4, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_it_goes_up_to_default_height(
            self, sleep_mock):
        # Fixture
        sut = PositionHlCommander(self.cf_mock, default_height=0.4, controller=PositionHlCommander.CONTROLLER_PID)

        # Test
        sut.take_off(velocity=0.6)

        # Assert
        duration = 0.4 / 0.6
        self.commander_mock.takeoff.assert_called_with(0.4, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_it_goes_down_on_landing(
            self, sleep_mock):
        # Fixture
        self.sut.take_off(height=0.4)

        # Test
        self.sut.land(velocity=0.6)

        # Assert
        duration = 0.4 / 0.6
        self.commander_mock.land.assert_called_with(0.0, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_it_goes_down_to_set_landing_height_on_landing(
            self, sleep_mock):
        # Fixture
        self.sut.take_off(height=1.0)
        self.sut.set_landing_height(0.4)

        # Test
        self.sut.land(velocity=0.6)

        # Assert
        duration = (1.0 - 0.4) / 0.6
        self.commander_mock.land.assert_called_with(0.4, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_it_takes_off_and_lands_as_context_manager(
            self, sleep_mock):
        # Fixture

        # Test
        with self.sut:
            pass

        # Assert
        duration1 = 0.5 / 0.5
        duration2 = 0.5 / 0.5
        self.commander_mock.takeoff.assert_called_with(0.5, duration1)
        self.commander_mock.land.assert_called_with(0.0, duration2)
        sleep_mock.assert_called_with(duration1)
        sleep_mock.assert_called_with(duration2)

    def test_that_it_returns_current_position(
            self, sleep_mock):
        # Fixture
        self.sut.take_off(height=0.4, velocity=0.6)

        # Test
        actual = self.sut.get_position()

        # Assert
        self.assertEqual(actual, (0.0, 0.0, 0.4))

    def test_that_it_goes_to_position(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()
        initial_pos = self.sut.get_position()

        # Test
        self.sut.go_to(1.0, 2.0, 3.0, 4.0)

        # Assert
        distance = self._distance(initial_pos, (1.0, 2.0, 3.0))
        duration = distance / 4.0
        self.commander_mock.go_to.assert_called_with(
            1.0, 2.0, 3.0, 0.0, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_it_does_not_send_goto_to_same_position(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()
        initial_pos = self.sut.get_position()

        # Test
        self.sut.go_to(initial_pos[0], initial_pos[1], initial_pos[2])

        # Assert
        self.commander_mock.go_to.assert_not_called()

    def test_that_it_moves_distance(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()
        initial_pos = self.sut.get_position()

        # Test
        self.sut.move_distance(1.0, 2.0, 3.0, 4.0)

        # Assert
        distance = self._distance((0.0, 0.0, 0.0), (1.0, 2.0, 3.0))
        duration = distance / 4.0
        final_pos = (
            initial_pos[0] + 1.0,
            initial_pos[1] + 2.0,
            initial_pos[2] + 3.0)
        self.commander_mock.go_to.assert_called_with(
            final_pos[0], final_pos[1], final_pos[2], 0.0, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_it_goes_forward(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()
        initial_pos = self.sut.get_position()

        # Test
        self.sut.forward(1.0, 2.0)

        # Assert
        duration = 1.0 / 2.0
        final_pos = (
            initial_pos[0] + 1.0,
            initial_pos[1],
            initial_pos[2])
        self.commander_mock.go_to.assert_called_with(
            final_pos[0], final_pos[1], final_pos[2], 0.0, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_it_goes_back(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()
        initial_pos = self.sut.get_position()

        # Test
        self.sut.back(1.0, 2.0)

        # Assert
        duration = 1.0 / 2.0
        final_pos = (
            initial_pos[0] - 1.0,
            initial_pos[1],
            initial_pos[2])
        self.commander_mock.go_to.assert_called_with(
            final_pos[0], final_pos[1], final_pos[2], 0.0, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_it_goes_left(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()
        initial_pos = self.sut.get_position()

        # Test
        self.sut.left(1.0, 2.0)

        # Assert
        duration = 1.0 / 2.0
        final_pos = (
            initial_pos[0],
            initial_pos[1] + 1.0,
            initial_pos[2])
        self.commander_mock.go_to.assert_called_with(
            final_pos[0], final_pos[1], final_pos[2], 0.0, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_it_goes_right(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()
        initial_pos = self.sut.get_position()

        # Test
        self.sut.right(1.0, 2.0)

        # Assert
        duration = 1.0 / 2.0
        final_pos = (
            initial_pos[0],
            initial_pos[1] - 1,
            initial_pos[2])
        self.commander_mock.go_to.assert_called_with(
            final_pos[0], final_pos[1], final_pos[2], 0, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_it_goes_up(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()
        initial_pos = self.sut.get_position()

        # Test
        self.sut.up(1.0, 2.0)

        # Assert
        duration = 1.0 / 2.0
        final_pos = (
            initial_pos[0],
            initial_pos[1],
            initial_pos[2] + 1)
        self.commander_mock.go_to.assert_called_with(
            final_pos[0], final_pos[1], final_pos[2], 0, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_it_goes_down(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()
        initial_pos = self.sut.get_position()

        # Test
        self.sut.down(1.0, 2.0)

        # Assert
        duration = 1.0 / 2.0
        final_pos = (
            initial_pos[0],
            initial_pos[1],
            initial_pos[2] - 1)
        self.commander_mock.go_to.assert_called_with(
            final_pos[0], final_pos[1], final_pos[2], 0, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_default_velocity_is_used(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()
        initial_pos = self.sut.get_position()
        self.sut.set_default_velocity(7)

        # Test
        self.sut.go_to(1.0, 2.0, 3.0)

        # Assert
        distance = self._distance(initial_pos, (1.0, 2.0, 3.0))
        duration = distance / 7.0
        self.commander_mock.go_to.assert_called_with(
            1.0, 2.0, 3.0, 0.0, duration)
        sleep_mock.assert_called_with(duration)

    def test_that_default_height_is_used(
            self, sleep_mock):
        # Fixture
        self.sut.take_off()
        initial_pos = self.sut.get_position()
        self.sut.set_default_velocity(7.0)
        self.sut.set_default_height(5.0)

        # Test
        self.sut.go_to(1.0, 2.0)

        # Assert
        distance = self._distance(initial_pos, (1.0, 2.0, 5.0))
        duration = distance / 7.0
        self.commander_mock.go_to.assert_called_with(
            1.0, 2.0, 5.0, 0.0, duration)
        sleep_mock.assert_called_with(duration)

    ######################################################################

    def _distance(self, p1, p2):
        dx = p1[0] - p2[0]
        dy = p1[1] - p2[1]
        dz = p1[2] - p2[2]
        return math.sqrt(dx * dx + dy * dy + dz * dz)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Used for sending control setpoints to the Crazyflie
"""
import struct

from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort

__author__ = 'Bitcraze AB'
__all__ = ['Commander']

TYPE_STOP = 0
TYPE_VELOCITY_WORLD = 1
TYPE_ZDISTANCE = 2
TYPE_HOVER = 5
TYPE_POSITION = 7


class Commander():
    """
    Used for sending control setpoints to the Crazyflie
    """

    def __init__(self, crazyflie=None):
        """
        Initialize the commander object. By default the commander is in
        +-mode (not x-mode).
        """
        self._cf = crazyflie
        self._x_mode = False

    def set_client_xmode(self, enabled):
        """
        Enable/disable the client side X-mode. When enabled this recalculates
        the setpoints before sending them to the Crazyflie.
        """
        self._x_mode = enabled

    def send_setpoint(self, roll, pitch, yawrate, thrust):
        """
        Send a new control setpoint for roll/pitch/yaw_Rate/thrust to the copter.

        The meaning of these values is depended on the mode of the RPYT commander in the firmware
        Default settings are Roll, pitch, yawrate and thrust

        roll,  pitch are in degrees
        yawrate is in degrees/s
        thrust is an integer value ranging from 10001 (next to no power) to 60000 (full power)
        """
        if thrust > 0xFFFF or thrust < 0:
            raise ValueError('Thrust must be between 0 and 0xFFFF')

        if self._x_mode:
            roll, pitch = 0.707 * (roll - pitch), 0.707 * (roll + pitch)

        pk = CRTPPacket()
        pk.port = CRTPPort.COMMANDER
        pk.data = struct.pack('<fffH', roll, -pitch, yawrate, thrust)
        self._cf.send_packet(pk)

    def send_stop_setpoint(self):
        """
        Send STOP setpoing, stopping the motors and (potentially) falling.
        """
        pk = CRTPPacket()
        pk.port = CRTPPort.COMMANDER_GENERIC
        pk.data = struct.pack('<B', TYPE_STOP)
        self._cf.send_packet(pk)

    def send_velocity_world_setpoint(self, vx, vy, vz, yawrate):
        """
        Send Velocity in the world frame of reference setpoint with yawrate commands

        vx, vy, vz are in m/s
        yawrate is in degrees/s
        """
        pk = CRTPPacket()
        pk.port = CRTPPort.COMMANDER_GENERIC
        pk.data = struct.pack('<Bffff', TYPE_VELOCITY_WORLD,
                              vx, vy, vz, yawrate)
        self._cf.send_packet(pk)

    def send_zdistance_setpoint(self, roll, pitch, yawrate, zdistance):
        """
        Control mode where the height is send as an absolute setpoint (intended
        to be the distance to the surface under the Crazflie), while giving roll,
        pitch and yaw rate commands

        roll, pitch are in degrees
        yawrate is in degrees/s
        zdistance is in meters
        """
        pk = CRTPPacket()
        pk.port = CRTPPort.COMMANDER_GENERIC
        pk.data = struct.pack('<Bffff', TYPE_ZDISTANCE,
                              roll, pitch, yawrate, zdistance)
        self._cf.send_packet(pk)

    def send_hover_setpoint(self, vx, vy, yawrate, zdistance):
        """
        Control mode where the height is send as an absolute setpoint (intended
        to be the distance to the surface under the Crazflie), while giving x, y velocity
        commands in body-fixed coordinates.

        vx,  vy are in m/s
        yawrate is in degrees/s
        zdistance is in meters
        """
        pk = CRTPPacket()
        pk.port = CRTPPort.COMMANDER_GENERIC
        pk.data = struct.pack('<Bffff', TYPE_HOVER,
                              vx, vy, yawrate, zdistance)
        self._cf.send_packet(pk)

    def send_position_setpoint(self, x, y, z, yaw):
        """
        Control mode where the position is sent as absolute (world) x,y,z coordinate in
        meter and the yaw is the absolute orientation.

        x, y, z are in m
        yaw is in degrees
        """
        pk = CRTPPacket()
        pk.port = CRTPPort.COMMANDER_GENERIC
        pk.data = struct.pack('<Bffff', TYPE_POSITION,
                              x, y, z, yaw)
        self._cf.send_packet(pk)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import unittest

import numpy as np
import numpy.typing as npt
from scipy.spatial.transform.rotation import Rotation

from cflib.localization.lighthouse_types import Pose


class LighthouseTestBase(unittest.TestCase):
    """
    Utilitis to simplify testing of lighthouse code
    """

    def assertVectorsAlmostEqual(self, expected: npt.ArrayLike, actual: npt.ArrayLike, places: int = 5) -> None:
        _expected = np.array(expected)
        _actual = np.array(actual)

        self.assertEqual(_expected.shape[0], _actual.shape[0], 'Shape differs')

        for i in range(_expected.shape[0]):
            self.assertAlmostEqual(_expected[i], _actual[i], places, f'Lists differs in element {i}')

    def assertPosesAlmostEqual(self, expected: Pose, actual: Pose, places: int = 5):
        translation_diff = expected.translation - actual.translation
        self.assertAlmostEqual(0.0, np.linalg.norm(translation_diff), places,
                               f'Translation different, expected: {expected.translation}, actual: {actual.translation}')

        def un_ambiguize(rot_vec):
            quat = Rotation.from_rotvec(rot_vec).as_quat()
            return Rotation.from_quat(quat).as_rotvec()

        _expected_rot_vec = un_ambiguize(expected.rot_vec)
        _actual_rot_vec = un_ambiguize(actual.rot_vec)

        rotation_diff = _expected_rot_vec - _actual_rot_vec
        self.assertAlmostEqual(0.0, np.linalg.norm(rotation_diff), places,
                               f'Rotation different, expected: {expected.rot_vec}, actual: {actual.rot_vec}')



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2016 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import struct
import unittest
from unittest.mock import MagicMock

from cflib.crazyflie import Crazyflie
from cflib.crazyflie.localization import Localization
from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort


class LocalizationTest(unittest.TestCase):

    def setUp(self):
        self.cf_mock = MagicMock(spec=Crazyflie)
        self.sut = Localization(crazyflie=self.cf_mock)

    def test_that_lighthouse_persist_data_is_correctly_encoded(self):

        # fixture
        geo_bs_list = [0, 2, 4, 6, 8, 10, 12, 14]
        calib_bs_list = [1, 3, 5, 7, 9, 11, 13, 15]

        # test
        actual = self.sut.send_lh_persist_data_packet(
            geo_bs_list, calib_bs_list)

        # assert
        data_check = 2863289685
        expected = CRTPPacket()
        expected.port = CRTPPort.LOCALIZATION
        expected.channel = self.sut.GENERIC_CH
        expected.data = struct.pack(
            '<BI', Localization.LH_PERSIST_DATA, data_check)

        actual_object = self.cf_mock.send_packet.call_args
        actual = actual_object[0][0]
        self.assertEqual(expected.port, actual.port)
        self.assertEqual(expected.channel, actual.channel)
        self.assertEqual(expected.data, actual.data)

    def test_that_lighthouse_persist_accepts_empty_lists(self):

        # fixture

        # tests and results
        self.sut.send_lh_persist_data_packet([], [])

        # Assert
        # The test should not raise an exception, so if we get here all is
        # good

    def test_that_lighthouse_persist_raises_for_geo_list_bs_too_high(self):
        # fixture
        max_bs_nr = 15
        invalid_list = [max_bs_nr + 1]

        # test
        # assert
        try:
            self.sut.send_lh_persist_data_packet(invalid_list, [])
        except Exception as e:
            actual = e.args[0]
            expected = 'Geometry BS list is not valid'
            self.assertEqual(expected, actual)
        else:
            self.fail('Expect exception')

    def test_that_lighthouse_persist_raises_for_calib_list_bs_too_high(self):
        # fixture
        max_bs_nr = 15
        invalid_list = [max_bs_nr + 1]

        # test
        # assert
        try:
            self.sut.send_lh_persist_data_packet([], invalid_list)
        except Exception as e:
            actual = e.args[0]
            expected = 'Calibration BS list is not valid'
            self.assertEqual(expected, actual)
        else:
            self.fail('Expect exception')

    def test_that_lighthouse_persist_raises_for_geo_list_bs_negative(self):
        # fixture
        invalid_list = [-1]

        # test
        # assert
        try:
            self.sut.send_lh_persist_data_packet(invalid_list, [])
        except Exception as e:
            actual = e.args[0]
            expected = 'Geometry BS list is not valid'
            self.assertEqual(expected, actual)
        else:
            self.fail('Expect exception')

    def test_that_lighthouse_persist_raises_for_calib_list_bs_negative(self):
        # fixture
        invalid_list = [-1]

        # test
        # assert
        try:
            self.sut.send_lh_persist_data_packet([], invalid_list)
        except Exception as e:
            actual = e.args[0]
            expected = 'Calibration BS list is not valid'
            self.assertEqual(expected, actual)
        else:
            self.fail('Expect exception')



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import time

from cflib.utils.power_switch import PowerSwitch


class RigSupport:
    def __init__(self):
        self.all_uris = [
            'radio://0/42/2M/E7E7E74201',
            'radio://0/42/2M/E7E7E74202',
            'radio://0/42/2M/E7E7E74203',
            'radio://0/42/2M/E7E7E74204',
            'radio://0/42/2M/E7E7E74205',
            'radio://0/42/2M/E7E7E74206',
            'radio://0/42/2M/E7E7E74207',
            'radio://0/42/2M/E7E7E74208',
            'radio://0/42/2M/E7E7E74209',
            'radio://0/42/2M/E7E7E7420A',
        ]

    def restart_devices(self, uris):
        print('Restarting devices')

        for uri in uris:
            PowerSwitch(uri).stm_power_down()

        time.sleep(1)

        for uri in uris:
            PowerSwitch(uri).stm_power_up()

        # Wait for devices to boot
        time.sleep(8)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""Scans and creates communication interfaces."""
import logging
import os

from .exceptions import WrongUriType
from .prrtdriver import PrrtDriver
from .radiodriver import RadioDriver
from .serialdriver import SerialDriver
from .tcpdriver import TcpDriver
from .udpdriver import UdpDriver
from .usbdriver import UsbDriver

__author__ = 'Bitcraze AB'
__all__ = []

logger = logging.getLogger(__name__)


CLASSES = []


def init_drivers(enable_debug_driver=False, enable_serial_driver=False):
    """Initialize all the drivers."""

    env = os.getenv('USE_CFLINK')
    if env is not None and env == 'cpp':
        from .cflinkcppdriver import CfLinkCppDriver
        CLASSES.append(CfLinkCppDriver)
    else:
        CLASSES.extend([RadioDriver, UsbDriver])

    if enable_debug_driver:
        logger.warn('The debug driver is no longer supported!')

    if enable_serial_driver:
        CLASSES.append(SerialDriver)

    CLASSES.extend([UdpDriver, PrrtDriver, TcpDriver])


def scan_interfaces(address=None):
    """ Scan all the interfaces for available Crazyflies """
    available = []
    found = []
    for driverClass in CLASSES:
        try:
            logger.debug('Scanning: %s', driverClass)
            instance = driverClass()
            found = instance.scan_interface(address)
            available += found
        except Exception:
            raise
    return available


def get_interfaces_status():
    """Get the status of all the interfaces"""
    status = {}
    for driverClass in CLASSES:
        try:
            instance = driverClass()
            status[instance.get_name()] = instance.get_status()
        except Exception:
            raise
    return status


def get_link_driver(uri, link_quality_callback=None, link_error_callback=None):
    """Return the link driver for the given URI. Returns None if no driver
    was found for the URI or the URI was not well formatted for the matching
    driver."""
    for driverClass in CLASSES:
        try:
            instance = driverClass()
            instance.connect(uri, link_quality_callback, link_error_callback)
            return instance
        except WrongUriType:
            continue

    return None



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import time
import unittest

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.swarm import CachedCfFactory
from cflib.crazyflie.swarm import Swarm
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from sys_test.swarm_test_rig.rig_support import RigSupport


class TestConnection(unittest.TestCase):
    def setUp(self):
        cflib.crtp.init_drivers()
        self.test_rig_support = RigSupport()

    def test_that_connection_time_scales_with_more_devices_without_cache(self):
        # Fixture
        self.test_rig_support.restart_devices(self.test_rig_support.all_uris)

        EXPECTED_CONNECTION_TIME = 5

        for nr_of_devices in range(1, len(self.test_rig_support.all_uris)):
            # Test
            uris = self.test_rig_support.all_uris[:nr_of_devices]

            start_time = time.time()
            with Swarm(uris):
                connected_time = time.time()

            actual = connected_time - start_time
            max_expected = EXPECTED_CONNECTION_TIME * nr_of_devices
            print('Connection time for', nr_of_devices, ':', actual,
                  ', per device:', actual / nr_of_devices)

            # Assert
            self.assertLess(actual, max_expected)

    def test_that_connection_time_scales_with_more_devices_with_cache(self):
        # Fixture
        self.test_rig_support.restart_devices(self.test_rig_support.all_uris)

        # Fill caches first by connecting to all devices
        factory = CachedCfFactory(rw_cache='./cache')
        with Swarm(self.test_rig_support.all_uris, factory=factory):
            pass

        EXPECTED_CONNECTION_TIME = 1.5

        for nr_of_devices in range(1, len(self.test_rig_support.all_uris)):
            # Test
            uris = self.test_rig_support.all_uris[:nr_of_devices]

            start_time = time.time()
            with Swarm(uris, factory=factory):
                connected_time = time.time()

            actual = connected_time - start_time
            max_expected = EXPECTED_CONNECTION_TIME * nr_of_devices
            print('Connection time for', nr_of_devices, ':', actual,
                  ', per device:', actual / nr_of_devices)

            # Assert
            self.assertLess(actual, max_expected)

    def test_that_all_devices_are_restarted(self):
        # Fixture
        uris = self.test_rig_support.all_uris

        # Test
        # Assert
        self.test_rig_support.restart_devices(uris)

    def test_that_all_devices_are_restarted_multiple_times(self):
        # Fixture
        uris = self.test_rig_support.all_uris

        # Test
        # Assert
        for i in range(10):
            self.test_rig_support.restart_devices(uris)

    def test_that_the_same_cf_object_can_be_connected_multiple_times(self):
        # Fixture
        self.test_rig_support.restart_devices(self.test_rig_support.all_uris)
        cf = Crazyflie(rw_cache='./cache')

        # Test
        for uri in self.test_rig_support.all_uris:
            with SyncCrazyflie(uri, cf=cf):
                pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2020 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Used for sending external position to the Crazyflie
"""

__author__ = 'Bitcraze AB'
__all__ = ['Extpos']


class Extpos():
    """
    Used for sending its position to the Crazyflie
    """

    def __init__(self, crazyflie=None):
        """
        Initialize the Extpos object.
        """
        self._cf = crazyflie

    def send_extpos(self, x, y, z):
        """
        Send the current Crazyflie X, Y, Z position. This is going to be
        forwarded to the Crazyflie's position estimator.
        """

        self._cf.loc.send_extpos([x, y, z])

    def send_extpose(self, x, y, z, qx, qy, qz, qw):
        """
        Send the current Crazyflie X, Y, Z position and attitude as a
        normalized quaternion. This is going to be forwarded to the
        Crazyflie's position estimator.
        """
        self._cf.loc.send_extpose([x, y, z], [qx, qy, qz, qw])



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import unittest

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.swarm import CachedCfFactory
from cflib.crazyflie.swarm import Swarm
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.crazyflie.syncLogger import SyncLogger
from sys_test.swarm_test_rig.rig_support import RigSupport


class TestLogging(unittest.TestCase):
    def setUp(self):
        cflib.crtp.init_drivers()
        self.test_rig_support = RigSupport()

    def test_that_requested_logging_is_received_properly_from_one_cf(self):
        # Fixture
        uri = self.test_rig_support.all_uris[0]
        self.test_rig_support.restart_devices([uri])
        cf = Crazyflie(rw_cache='./cache')

        # Test and Assert
        with SyncCrazyflie(uri, cf=cf) as scf:
            self.assert_add_logging_and_get_non_zero_value(scf)

    def test_that_requested_logging_is_received_properly_from_all_cfs(self):
        # Fixture
        uris = self.test_rig_support.all_uris
        self.test_rig_support.restart_devices(uris)
        factory = CachedCfFactory(rw_cache='./cache')

        # Test and Assert
        with Swarm(uris, factory=factory) as swarm:
            swarm.parallel_safe(self.assert_add_logging_and_get_non_zero_value)

    def assert_add_logging_and_get_non_zero_value(self, scf):
        log_name = 'stabilizer.roll'
        expected = 0.0

        lg_conf = LogConfig(name='SysTest', period_in_ms=10)
        lg_conf.add_variable(log_name, 'float')

        with SyncLogger(scf, lg_conf) as logger:
            for log_entry in logger:
                actual = log_entry[1][log_name]
                break

        self.assertNotAlmostEqual(expected, actual, places=4)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2021 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Functionality to manage lighthouse system configuration (geometry and calibration data).
"""
import time

import yaml

from cflib.crazyflie.mem import LighthouseBsCalibration
from cflib.crazyflie.mem import LighthouseBsGeometry
from cflib.crazyflie.mem import LighthouseMemHelper


class LighthouseConfigFileManager:
    TYPE_ID = 'type'
    TYPE = 'lighthouse_system_configuration'
    VERSION_ID = 'version'
    VERSION = '1'
    GEOS_ID = 'geos'
    CALIBS_ID = 'calibs'
    SYSTEM_TYPE_ID = 'systemType'

    SYSTEM_TYPE_V1 = 1
    SYSTEM_TYPE_V2 = 2

    @staticmethod
    def write(file_name, geos={}, calibs={}, system_type=SYSTEM_TYPE_V2):
        file = open(file_name, 'w')
        with file:
            file_geos = {}
            for id, geo in geos.items():
                if geo.valid:
                    file_geos[id] = geo.as_file_object()

            file_calibs = {}
            for id, calib in calibs.items():
                if calib.valid:
                    file_calibs[id] = calib.as_file_object()

            data = {
                LighthouseConfigFileManager.TYPE_ID: LighthouseConfigFileManager.TYPE,
                LighthouseConfigFileManager.VERSION_ID: LighthouseConfigFileManager.VERSION,
                LighthouseConfigFileManager.SYSTEM_TYPE_ID: system_type,
                LighthouseConfigFileManager.GEOS_ID: file_geos,
                LighthouseConfigFileManager.CALIBS_ID: file_calibs
            }

            yaml.dump(data, file)

    @staticmethod
    def read(file_name):
        file = open(file_name, 'r')
        with file:
            data = yaml.safe_load(file)

            if LighthouseConfigFileManager.TYPE_ID not in data:
                raise Exception('Type field missing')

            if data[LighthouseConfigFileManager.TYPE_ID] != LighthouseConfigFileManager.TYPE:
                raise Exception('Unsupported file type')

            if LighthouseConfigFileManager.VERSION_ID not in data:
                raise Exception('Version field missing')

            if data[LighthouseConfigFileManager.VERSION_ID] != LighthouseConfigFileManager.VERSION:
                raise Exception('Unsupported file version')

            result_system_type = LighthouseConfigFileManager.SYSTEM_TYPE_V2
            if LighthouseConfigFileManager.SYSTEM_TYPE_ID in data:
                result_system_type = data[LighthouseConfigFileManager.SYSTEM_TYPE_ID]

            result_geos = {}
            result_calibs = {}

            if LighthouseConfigFileManager.GEOS_ID in data:
                for id, geo in data[LighthouseConfigFileManager.GEOS_ID].items():
                    result_geos[id] = LighthouseBsGeometry.from_file_object(geo)

            if LighthouseConfigFileManager.CALIBS_ID in data:
                for id, calib in data[LighthouseConfigFileManager.CALIBS_ID].items():
                    result_calibs[id] = LighthouseBsCalibration.from_file_object(calib)

            return result_geos, result_calibs, result_system_type


class LighthouseConfigWriter:
    """
    This class is used to write system config data to the Crazyflie RAM and persis to permanent storage
    """

    def __init__(self, cf, nr_of_base_stations=16):
        self._cf = cf
        self._helper = LighthouseMemHelper(cf)
        self._data_stored_cb = None
        self._geos_to_write = None
        self._geos_to_persist = []
        self._calibs_to_persist = []
        self._write_failed_for_one_or_more_objects = False
        self._nr_of_base_stations = nr_of_base_stations

    def write_and_store_config(self, data_stored_cb, geos=None, calibs=None, system_type=None):
        """
        Transfer geometry and calibration data to the Crazyflie and persist to permanent storage.
        The callback is called when done.
        If geos or calibs is None, no data will be written for that data type.
        If geos or calibs is a dictionary, the values for the base stations in the dictionary will
        transfered to the Crazyflie, data for all other base stations will be invalidated.
        """
        if self._data_stored_cb is not None:
            raise Exception('Write already in prgress')
        self._data_stored_cb = data_stored_cb

        self._cf.loc.receivedLocationPacket.add_callback(self._received_location_packet)

        self._geos_to_write = self._prepare_geos(geos)
        self._calibs_to_write = self._prepare_calibs(calibs)

        self._geos_to_persist = []
        if self._geos_to_write is not None:
            self._geos_to_persist = list(range(self._nr_of_base_stations))

        self._calibs_to_persist = []
        if self._calibs_to_write is not None:
            self._calibs_to_persist = list(range(self._nr_of_base_stations))

        self._write_failed_for_one_or_more_objects = False

        if system_type is not None:
            # Change system type first as this will erase calib and geo data in the CF.
            # Changing system type may trigger a lengthy operation (up to 0.5 s) if the persistant memory requires
            # defrag. Setting a param is an asynchronous operataion, and it is not possible to know if the system
            # swich is finished before we continue.
            self._cf.param.set_value('lighthouse.systemType', system_type)

            # We add a sleep here to make sure the change of system type is finished. It is dirty but will have to
            # do for now. A more propper solution would be to add support for Remote Procedure Calls (RPC) with
            # synchronous function calls.
            time.sleep(0.8)

        self._next()

    def write_and_store_config_from_file(self, data_stored_cb, file_name):
        """
        Read system configuration data from file and write/persist to the Crazyflie.
        Geometry and calibration data for base stations that are not in the config file will be invalidated.
        """
        geos, calibs, system_type = LighthouseConfigFileManager.read(file_name)
        self.write_and_store_config(data_stored_cb, geos=geos, calibs=calibs, system_type=system_type)

    def _next(self):
        if self._geos_to_write is not None:
            self._helper.write_geos(self._geos_to_write, self._upload_done)
            self._geos_to_write = None
            return

        if self._calibs_to_write is not None:
            self._helper.write_calibs(self._calibs_to_write, self._upload_done)
            self._calibs_to_write = None
            return

        if len(self._geos_to_persist) > 0 or len(self._calibs_to_persist) > 0:
            self._cf.loc.send_lh_persist_data_packet(self._geos_to_persist, self._calibs_to_persist)
            self._geos_to_persist = []
            self._calibs_to_persist = []
            return

        tmp_callback = self._data_stored_cb
        self._data_stored_cb = None
        if tmp_callback is not None:
            tmp_callback(not self._write_failed_for_one_or_more_objects)

    def _upload_done(self, sucess):
        if not sucess:
            self._write_failed_for_one_or_more_objects = True
        self._next()

    def _received_location_packet(self, packet):
        # New geo data has been written and stored in the CF
        if packet.type == self._cf.loc.LH_PERSIST_DATA:
            self._next()

    def _prepare_geos(self, geos):
        result = None

        if geos is not None:
            result = dict(geos)

            # Pad for base stations without data
            empty_geo = LighthouseBsGeometry()
            for id in range(self._nr_of_base_stations):
                if id not in result:
                    result[id] = empty_geo

        return result

    def _prepare_calibs(self, calibs):
        result = None

        if calibs is not None:
            result = dict(calibs)

            # Pad for base stations without data
            empty_calib = LighthouseBsCalibration()
            for id in range(self._nr_of_base_stations):
                if id not in result:
                    result[id] = empty_calib

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2021 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>
import os
import sys


def uri_from_env(env='CFLIB_URI', default='radio://0/80/2M/E7E7E7E7E7') -> str:
    try:
        return os.environ[env]
    except KeyError:
        return default


def address_from_env(env='CFLIB_URI', default=0xE7E7E7E7E7) -> int:
    try:
        uri = os.environ[env]
    except KeyError:
        return default

    # Get the address part of the uri
    address = uri.rsplit('/', 1)[-1]
    try:
        return int(address, 16)
    except ValueError:
        print('address is not hexadecimal! (%s)' % address, file=sys.stderr)
        return None



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2021 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import unittest

from cflib.crazyflie.mem import LighthouseBsCalibration
from cflib.crazyflie.mem.lighthouse_memory import LighthouseBsGeometry


class TestLighthouseMemory(unittest.TestCase):
    def test_bs_calibration_file_format(self):
        # Fixture
        calib = LighthouseBsCalibration()
        calib.uid = 1234

        calib.sweeps[0].curve = 1.0
        calib.sweeps[0].phase = 2.0
        calib.sweeps[0].tilt = 3.0
        calib.sweeps[0].gibmag = 4.0
        calib.sweeps[0].gibphase = 5.0
        calib.sweeps[0].ogeemag = 6.0
        calib.sweeps[0].ogeephase = 7.0

        calib.sweeps[1].curve = 8.0

        # Test
        actual = calib.as_file_object()

        # Assert
        self.assertEqual(1234, actual['uid'])
        self.assertEqual(1.0, actual['sweeps'][0]['curve'])
        self.assertEqual(2.0, actual['sweeps'][0]['phase'])
        self.assertEqual(3.0, actual['sweeps'][0]['tilt'])
        self.assertEqual(4.0, actual['sweeps'][0]['gibmag'])
        self.assertEqual(5.0, actual['sweeps'][0]['gibphase'])
        self.assertEqual(6.0, actual['sweeps'][0]['ogeemag'])
        self.assertEqual(7.0, actual['sweeps'][0]['ogeephase'])
        self.assertEqual(8.0, actual['sweeps'][1]['curve'])

    def test_bs_calibration_file_write_read(self):
        # Fixture
        calib = LighthouseBsCalibration()
        calib.uid = 1234

        calib.sweeps[0].curve = 1.0
        calib.sweeps[0].phase = 2.0
        calib.sweeps[0].tilt = 3.0
        calib.sweeps[0].gibmag = 4.0
        calib.sweeps[0].gibphase = 5.0
        calib.sweeps[0].ogeemag = 6.0
        calib.sweeps[0].ogeephase = 7.0

        calib.sweeps[1].curve = 8.0

        file_object = calib.as_file_object()

        # Test
        actual = LighthouseBsCalibration.from_file_object(file_object)

        # Assert
        actual_file_object = actual.as_file_object()
        self.assertEqual(file_object, actual_file_object)
        self.assertTrue(actual.valid)

    def test_bs_geometry_file_format(self):
        # Fixture
        geo = LighthouseBsGeometry()
        geo.origin = [1.0, 2.0, 3.0]
        geo.rotation_matrix = [[1.0, 2.0, 3.0], [1.1, 2.1, 3.1], [1.2, 2.2, 3.2]]

        # Test
        actual = geo.as_file_object()

        # Assert
        self.assertEqual([1.0, 2.0, 3.0], actual['origin'])
        self.assertEqual([[1.0, 2.0, 3.0], [1.1, 2.1, 3.1], [1.2, 2.2, 3.2]], actual['rotation'])

    def test_bs_geometry_file_write_read(self):
        # Fixture
        geo = LighthouseBsGeometry()
        geo.origin = [1.0, 2.0, 3.0]
        geo.rotation_matrix = [[1.0, 2.0, 3.0], [1.1, 2.1, 3.1], [1.2, 2.2, 3.2]]

        file_object = geo.as_file_object()

        # Test
        actual = LighthouseBsGeometry.from_file_object(file_object)

        # Assert
        actual_file_object = actual.as_file_object()
        self.assertEqual(file_object, actual_file_object)
        self.assertTrue(actual.valid)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
The Crazyflie module is used to easily connect/send/receive data
from a Crazyflie.

Each function in the Crazyflie has a class in the module that can be used
to access that functionality. The same design is then used in the Crazyflie
firmware which makes the mapping 1:1 in most cases.
"""
import datetime
import logging
import time
from collections import namedtuple
from threading import current_thread
from threading import Lock
from threading import Thread
from threading import Timer

import cflib.crtp
from .appchannel import Appchannel
from .commander import Commander
from .console import Console
from .extpos import Extpos
from .localization import Localization
from .log import Log
from .mem import Memory
from .param import Param
from .platformservice import PlatformService
from .toccache import TocCache
from cflib.crazyflie.high_level_commander import HighLevelCommander
from cflib.utils.callbacks import Caller

__author__ = 'Bitcraze AB'
__all__ = ['Crazyflie']

logger = logging.getLogger(__name__)


class State:
    """Stat of the connection procedure"""
    DISCONNECTED = 0
    INITIALIZED = 1
    CONNECTED = 2
    SETUP_FINISHED = 3


class Crazyflie():
    """The Crazyflie class"""

    def __init__(self, link=None, ro_cache=None, rw_cache=None):
        """
        Create the objects from this module and register callbacks.

        @param ro_cache Path to read-only cache (string)
        @param rw_cache Path to read-write cache (string)
        """

        # Called on disconnect, no matter the reason
        self.disconnected = Caller()
        # Called on unintentional disconnect only
        self.connection_lost = Caller()
        # Called when the first packet in a new link is received
        self.link_established = Caller()
        # Called when the user requests a connection
        self.connection_requested = Caller()
        # Called when the link is established and the TOCs (that are not cached) have been downloaded
        self.connected = Caller()
        # Called when the the link is established and all data, including parameters have been downloaded
        self.fully_connected = Caller()

        # Called if establishing of the link fails (i.e times out)
        self.connection_failed = Caller()
        # Called for every packet received
        self.packet_received = Caller()
        # Called for every packet sent
        self.packet_sent = Caller()
        # Called when the link driver updates the link quality measurement
        self.link_quality_updated = Caller()

        self.state = State.DISCONNECTED

        self.link = link
        self._toc_cache = TocCache(ro_cache=ro_cache,
                                   rw_cache=rw_cache)

        self.incoming = _IncomingPacketHandler(self)
        self.incoming.setDaemon(True)
        if self.link:
            self.incoming.start()

        self.commander = Commander(self)
        self.high_level_commander = HighLevelCommander(self)
        self.loc = Localization(self)
        self.extpos = Extpos(self)
        self.log = Log(self)
        self.console = Console(self)
        self.param = Param(self)
        self.mem = Memory(self)
        self.platform = PlatformService(self)
        self.appchannel = Appchannel(self)

        self.link_uri = ''

        # Used for retry when no reply was sent back
        self.packet_received.add_callback(self._check_for_initial_packet_cb)
        self.packet_received.add_callback(self._check_for_answers)

        self._answer_patterns = {}

        self._send_lock = Lock()

        self.connected_ts = None

        self.param.all_updated.add_callback(self._all_parameters_updated)

        # Connect callbacks to logger
        self.disconnected.add_callback(
            lambda uri: logger.info('Callback->Disconnected from [%s]', uri))
        self.disconnected.add_callback(self._disconnected)
        self.link_established.add_callback(
            lambda uri: logger.info('Callback->Connected to [%s]', uri))
        self.connection_lost.add_callback(
            lambda uri, errmsg: logger.info('Callback->Connection lost to [%s]: %s', uri, errmsg))
        self.connection_failed.add_callback(
            lambda uri, errmsg: logger.info('Callback->Connected failed to [%s]: %s', uri, errmsg))
        self.connection_requested.add_callback(
            lambda uri: logger.info('Callback->Connection initialized[%s]', uri))
        self.connected.add_callback(
            lambda uri: logger.info('Callback->Connection setup finished [%s]', uri))
        self.fully_connected.add_callback(
            lambda uri: logger.info('Callback->Connection completed [%s]', uri))

    def _disconnected(self, link_uri):
        """ Callback when disconnected."""
        self.connected_ts = None

    def _start_connection_setup(self):
        """Start the connection setup by refreshing the TOCs"""
        logger.info('We are connected[%s], request connection setup',
                    self.link_uri)
        self.platform.fetch_platform_informations(self._platform_info_fetched)

    def _platform_info_fetched(self):
        self.log.refresh_toc(self._log_toc_updated_cb, self._toc_cache)

    def _param_toc_updated_cb(self):
        """Called when the param TOC has been fully updated"""
        logger.info('Param TOC finished updating')
        self.connected_ts = datetime.datetime.now()
        self.connected.call(self.link_uri)
        # Trigger the update for all the parameters
        self.param.request_update_of_all_params()

    def _mems_updated_cb(self):
        """Called when the memories have been identified"""
        logger.info('Memories finished updating')
        self.param.refresh_toc(self._param_toc_updated_cb, self._toc_cache)

    def _log_toc_updated_cb(self):
        """Called when the log TOC has been fully updated"""
        logger.info('Log TOC finished updating')
        self.mem.refresh(self._mems_updated_cb)

    def _all_parameters_updated(self):
        """Called when all parameters have been updated"""
        logger.info('All parameters updated')
        self.fully_connected.call(self.link_uri)

    def _link_error_cb(self, errmsg):
        """Called from the link driver when there's an error"""
        logger.warning('Got link error callback [%s] in state [%s]',
                       errmsg, self.state)
        if (self.link is not None):
            self.link.close()
        self.link = None
        if (self.state == State.INITIALIZED):
            self.connection_failed.call(self.link_uri, errmsg)
        if (self.state == State.CONNECTED or
                self.state == State.SETUP_FINISHED):
            self.disconnected.call(self.link_uri)
            self.connection_lost.call(self.link_uri, errmsg)
        self.state = State.DISCONNECTED

    def _link_quality_cb(self, percentage):
        """Called from link driver to report link quality"""
        self.link_quality_updated.call(percentage)

    def _check_for_initial_packet_cb(self, data):
        """
        Called when first packet arrives from Crazyflie.

        This is used to determine if we are connected to something that is
        answering.
        """
        self.state = State.CONNECTED
        self.link_established.call(self.link_uri)
        self.packet_received.remove_callback(self._check_for_initial_packet_cb)

    def open_link(self, link_uri):
        """
        Open the communication link to a copter at the given URI and setup the
        connection (download log/parameter TOC).
        """
        self.connection_requested.call(link_uri)
        self.state = State.INITIALIZED
        self.link_uri = link_uri
        try:
            self.link = cflib.crtp.get_link_driver(
                link_uri, self._link_quality_cb, self._link_error_cb)

            if not self.link:
                message = 'No driver found or malformed URI: {}' \
                    .format(link_uri)
                logger.warning(message)
                self.connection_failed.call(link_uri, message)
            else:
                if not self.incoming.is_alive():
                    self.incoming.start()
                # Add a callback so we can check that any data is coming
                # back from the copter
                self.packet_received.add_callback(
                    self._check_for_initial_packet_cb)

                self._start_connection_setup()
        except Exception as ex:  # pylint: disable=W0703
            # We want to catch every possible exception here and show
            # it in the user interface
            import traceback

            logger.error("Couldn't load link driver: %s\n\n%s",
                         ex, traceback.format_exc())
            exception_text = "Couldn't load link driver: %s\n\n%s" % (
                ex, traceback.format_exc())
            if self.link:
                self.link.close()
                self.link = None
            self.connection_failed.call(link_uri, exception_text)

    def close_link(self):
        """Close the communication link."""
        logger.info('Closing link')
        if (self.link is not None):
            self.commander.send_setpoint(0, 0, 0, 0)
        if (self.link is not None):
            self.link.close()
            self.link = None
        self._answer_patterns = {}
        self.disconnected.call(self.link_uri)
        self.state = State.DISCONNECTED

    """Check if the communication link is open or not."""

    def is_connected(self):
        return self.connected_ts is not None

    def add_port_callback(self, port, cb):
        """Add a callback to cb on port"""
        self.incoming.add_port_callback(port, cb)

    def remove_port_callback(self, port, cb):
        """Remove the callback cb on port"""
        self.incoming.remove_port_callback(port, cb)

    def _no_answer_do_retry(self, pk, pattern):
        """Resend packets that we have not gotten answers to"""
        logger.info('Resending for pattern %s', pattern)
        # Set the timer to None before trying to send again
        self.send_packet(pk, expected_reply=pattern, resend=True)

    def _check_for_answers(self, pk):
        """
        Callback called for every packet received to check if we are
        waiting for an answer on this port. If so, then cancel the retry
        timer.
        """
        longest_match = ()
        if len(self._answer_patterns) > 0:
            data = (pk.header,) + tuple(pk.data)
            for p in list(self._answer_patterns.keys()):
                logger.debug('Looking for pattern match on %s vs %s', p, data)
                if len(p) <= len(data):
                    if p == data[0:len(p)]:
                        match = data[0:len(p)]
                        if len(match) >= len(longest_match):
                            logger.debug('Found new longest match %s', match)
                            longest_match = match
        if len(longest_match) > 0:
            self._answer_patterns[longest_match].cancel()
            del self._answer_patterns[longest_match]

    def send_packet(self, pk, expected_reply=(), resend=False, timeout=0.2):
        """
        Send a packet through the link interface.

        @param pk Packet to send
        @param expect_answer True if a packet from the Crazyflie is expected to
                             be sent back, otherwise false

        """

        if not pk.is_data_size_valid():
            raise Exception('Data part of packet is too large')

        self._send_lock.acquire()
        if self.link is not None:
            if len(expected_reply) > 0 and not resend and \
                    self.link.needs_resending:
                pattern = (pk.header,) + expected_reply
                logger.debug(
                    'Sending packet and expecting the %s pattern back',
                    pattern)
                new_timer = Timer(timeout,
                                  lambda: self._no_answer_do_retry(pk,
                                                                   pattern))
                self._answer_patterns[pattern] = new_timer
                new_timer.start()
            elif resend:
                # Check if we have gotten an answer, if not try again
                pattern = expected_reply
                if pattern in self._answer_patterns:
                    logger.debug('We want to resend and the pattern is there')
                    if self._answer_patterns[pattern]:
                        new_timer = Timer(timeout,
                                          lambda:
                                          self._no_answer_do_retry(
                                              pk, pattern))
                        self._answer_patterns[pattern] = new_timer
                        new_timer.start()
                else:
                    logger.debug('Resend requested, but no pattern found: %s',
                                 self._answer_patterns)
            self.link.send_packet(pk)
            self.packet_sent.call(pk)
        self._send_lock.release()

    def is_called_by_incoming_handler_thread(self):
        return current_thread() == self.incoming


_CallbackContainer = namedtuple('CallbackConstainer',
                                'port port_mask channel channel_mask callback')


class _IncomingPacketHandler(Thread):
    """Handles incoming packets and sends the data to the correct receivers"""

    def __init__(self, cf):
        Thread.__init__(self)
        self.cf = cf
        self.cb = []

    def add_port_callback(self, port, cb):
        """Add a callback for data that comes on a specific port"""
        logger.debug('Adding callback on port [%d] to [%s]', port, cb)
        self.add_header_callback(cb, port, 0, 0xff, 0x0)

    def remove_port_callback(self, port, cb):
        """Remove a callback for data that comes on a specific port"""
        logger.debug('Removing callback on port [%d] to [%s]', port, cb)
        for port_callback in self.cb:
            if port_callback.port == port and port_callback.callback == cb:
                self.cb.remove(port_callback)

    def add_header_callback(self, cb, port, channel, port_mask=0xFF,
                            channel_mask=0xFF):
        """
        Add a callback for a specific port/header callback with the
        possibility to add a mask for channel and port for multiple
        hits for same callback.
        """
        self.cb.append(_CallbackContainer(port, port_mask,
                                          channel, channel_mask, cb))

    def run(self):
        while True:
            if self.cf.link is None:
                time.sleep(1)
                continue
            pk = self.cf.link.receive_packet(1)

            if pk is None:
                continue

            # All-packet callbacks
            self.cf.packet_received.call(pk)

            found = False
            for cb in (cb for cb in self.cb
                       if cb.port == (pk.port & cb.port_mask) and
                       cb.channel == (pk.channel & cb.channel_mask)):
                try:
                    cb.callback(pk)
                except Exception:  # pylint: disable=W0703
                    # Disregard pylint warning since we want to catch all
                    # exceptions and we can't know what will happen in
                    # the callbacks.
                    import traceback

                    logger.error('Exception while doing callback on port'
                                 ' [%d]\n\n%s', pk.port,
                                 traceback.format_exc())
                if cb.port != 0xFF:
                    found = True

            if not found:
                pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
A generic TableOfContents module that is used to fetch, store and manipulate
a TOC for logging or parameters.
"""
import logging
import struct

from cflib.crtp.crtpstack import CRTPPacket

__author__ = 'Bitcraze AB'
__all__ = ['Toc', 'TocFetcher']

logger = logging.getLogger(__name__)

TOC_CHANNEL = 0

# Commands used when accessing the Table of Contents
CMD_TOC_ELEMENT = 0  # original version: up to 255 entries
CMD_TOC_INFO = 1    # original version: up to 255 entries
CMD_TOC_ITEM_V2 = 2  # version 2: up to 16k entries
CMD_TOC_INFO_V2 = 3  # version 2: up to 16k entries

# Possible states when receiving TOC
IDLE = 'IDLE'
GET_TOC_INFO = 'GET_TOC_INFO'
GET_TOC_ELEMENT = 'GET_TOC_ELEMENT'


class Toc:
    """Container for TocElements."""

    def __init__(self):
        self.toc = {}

    def clear(self):
        """Clear the TOC"""
        self.toc = {}

    def add_element(self, element):
        """Add a new TocElement to the TOC container."""
        try:
            self.toc[element.group][element.name] = element
        except KeyError:
            self.toc[element.group] = {}
            self.toc[element.group][element.name] = element

    def get_element_by_complete_name(self, complete_name):
        """Get a TocElement element identified by complete name from the
        container."""
        try:
            return self.get_element_by_id(self.get_element_id(complete_name))
        except ValueError:
            # Item not found
            return None

    def get_element_id(self, complete_name):
        """Get the TocElement element id-number of the element with the
        supplied name."""
        [group, name] = complete_name.split('.')
        element = self.get_element(group, name)
        if element:
            return element.ident
        else:
            logger.warning('Unable to find variable [%s]', complete_name)
            return None

    def get_element(self, group, name):
        """Get a TocElement element identified by name and group from the
        container."""
        try:
            return self.toc[group][name]
        except KeyError:
            return None

    def get_element_by_id(self, ident):
        """Get a TocElement element identified by index number from the
        container."""
        for group in list(self.toc.keys()):
            for name in list(self.toc[group].keys()):
                if self.toc[group][name].ident == ident:
                    return self.toc[group][name]
        return None


class TocFetcher:
    """Fetches TOC entries from the Crazyflie"""

    def __init__(self, crazyflie, element_class, port, toc_holder,
                 finished_callback, toc_cache):
        self.cf = crazyflie
        self.port = port
        self._crc = 0
        self.requested_index = None
        self.nbr_of_items = None
        self.state = None
        self.toc = toc_holder
        self._toc_cache = toc_cache
        self.finished_callback = finished_callback
        self.element_class = element_class
        self._useV2 = False

    def start(self):
        """Initiate fetching of the TOC."""
        self._useV2 = self.cf.platform.get_protocol_version() >= 4

        logger.debug('[%d]: Using V2 protocol: %d', self.port, self._useV2)

        logger.debug('[%d]: Start fetching...', self.port)
        # Register callback in this class for the port
        self.cf.add_port_callback(self.port, self._new_packet_cb)

        # Request the TOC CRC
        self.state = GET_TOC_INFO
        pk = CRTPPacket()
        pk.set_header(self.port, TOC_CHANNEL)
        if self._useV2:
            pk.data = (CMD_TOC_INFO_V2,)
            self.cf.send_packet(pk, expected_reply=(CMD_TOC_INFO_V2,))
        else:
            pk.data = (CMD_TOC_INFO,)
            self.cf.send_packet(pk, expected_reply=(CMD_TOC_INFO,))

    def _toc_fetch_finished(self):
        """Callback for when the TOC fetching is finished"""
        self.cf.remove_port_callback(self.port, self._new_packet_cb)
        logger.debug('[%d]: Done!', self.port)
        self.finished_callback()

    def _new_packet_cb(self, packet):
        """Handle a newly arrived packet"""
        chan = packet.channel
        if (chan != 0):
            return
        payload = packet.data[1:]

        if (self.state == GET_TOC_INFO):
            if self._useV2:
                [self.nbr_of_items, self._crc] = struct.unpack(
                    '<HI', payload[:6])
            else:
                [self.nbr_of_items, self._crc] = struct.unpack(
                    '<BI', payload[:5])
            logger.debug('[%d]: Got TOC CRC, %d items and crc=0x%08X',
                         self.port, self.nbr_of_items, self._crc)

            cache_data = self._toc_cache.fetch(self._crc)
            if (cache_data):
                self.toc.toc = cache_data
                logger.info('TOC for port [%s] found in cache' % self.port)
                self._toc_fetch_finished()
            else:
                self.state = GET_TOC_ELEMENT
                self.requested_index = 0
                if (self.nbr_of_items > 0):
                    self._request_toc_element(self.requested_index)
                else:
                    logger.debug('No TOC entries for port [%s]' % self.port)
                    self._toc_cache.insert(self._crc, self.toc.toc)
                    self._toc_fetch_finished()

        elif (self.state == GET_TOC_ELEMENT):
            # Always add new element, but only request new if it's not the
            # last one.
            if self._useV2:
                ident = struct.unpack('<H', payload[:2])[0]
            else:
                ident = payload[0]

            if ident != self.requested_index:
                return
            if self._useV2:
                self.toc.add_element(self.element_class(ident, payload[2:]))
            else:
                self.toc.add_element(self.element_class(ident, payload[1:]))
            logger.debug('Added element [%s]', ident)
            if (self.requested_index < (self.nbr_of_items - 1)):
                logger.debug('[%d]: More variables, requesting index %d',
                             self.port, self.requested_index + 1)
                self.requested_index += 1
                self._request_toc_element(self.requested_index)
            else:  # No more variables in TOC
                self._toc_cache.insert(self._crc, self.toc.toc)
                self._toc_fetch_finished()

    def _request_toc_element(self, index):
        """Request information about a specific item in the TOC"""
        logger.debug('Requesting index %d on port %d', index, self.port)
        pk = CRTPPacket()
        if self._useV2:
            pk.set_header(self.port, TOC_CHANNEL)
            pk.data = (CMD_TOC_ITEM_V2, index & 0x0ff, (index >> 8) & 0x0ff)
            self.cf.send_packet(pk, expected_reply=(
                CMD_TOC_ITEM_V2, index & 0x0ff, (index >> 8) & 0x0ff))
        else:
            pk.set_header(self.port, TOC_CHANNEL)
            pk.data = (CMD_TOC_ELEMENT, index)
            self.cf.send_packet(pk, expected_reply=(CMD_TOC_ELEMENT, index))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Access the TOC cache for reading/writing. It supports both user
cache and dist cache.
"""
import json
import logging
import os
from glob import glob

from .log import LogTocElement  # noqa
from .param import ParamTocElement  # noqa

__author__ = 'Bitcraze AB'
__all__ = ['TocCache']

logger = logging.getLogger(__name__)


class TocCache():
    """
    Access to TOC cache. To turn of the cache functionality
    don't supply any directories.
    """

    def __init__(self, ro_cache=None, rw_cache=None):
        self._cache_files = []
        if (ro_cache):
            self._cache_files += glob(ro_cache + '/*.json')
        if (rw_cache):
            self._cache_files += glob(rw_cache + '/*.json')
            if not os.path.exists(rw_cache):
                os.makedirs(rw_cache)

        self._rw_cache = rw_cache

    def fetch(self, crc):
        """ Try to get a hit in the cache, return None otherwise """
        cache_data = None
        pattern = '%08X.json' % crc
        hit = None

        for name in self._cache_files:
            if (name.endswith(pattern)):
                hit = name

        if (hit):
            try:
                cache = open(hit)
                cache_data = json.load(cache,
                                       object_hook=self._decoder)
                cache.close()
            except Exception as exp:
                logger.warning('Error while parsing cache file [%s]:%s',
                               hit, str(exp))

        return cache_data

    def insert(self, crc, toc):
        """ Save a new cache to file """
        if self._rw_cache:
            try:
                filename = '%s/%08X.json' % (self._rw_cache, crc)
                cache = open(filename, 'w')
                cache.write(json.dumps(toc, indent=2,
                                       default=self._encoder))
                cache.close()
                logger.info('Saved cache to [%s]', filename)
                self._cache_files += [filename]
            except Exception as exp:
                logger.warning('Could not save cache to file [%s]: %s',
                               filename, str(exp))
        else:
            logger.warning('Could not save cache, no writable directory')

    def _encoder(self, obj):
        """ Encode a toc element leaf-node """
        encoded = {
            '__class__': obj.__class__.__name__,
            'ident': obj.ident,
            'group': obj.group,
            'name': obj.name,
            'ctype': obj.ctype,
            'pytype': obj.pytype,
            'access': obj.access,
        }

        if isinstance(obj, ParamTocElement):
            encoded['extended'] = obj.extended

        return encoded
        raise TypeError(repr(obj) + ' is not JSON serializable')

    def _decoder(self, obj):
        """ Decode a toc element leaf-node """
        if '__class__' in obj:
            elem = eval(obj['__class__'])()
            elem.ident = obj['ident']
            elem.group = str(obj['group'])
            elem.name = str(obj['name'])
            elem.ctype = str(obj['ctype'])
            elem.pytype = str(obj['pytype'])
            elem.access = obj['access']
            if isinstance(elem, ParamTocElement):
                elem.extended = obj['extended']
            return elem
        return obj



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2016 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import unittest
from unittest.mock import MagicMock

from cflib.crazyflie.swarm import Swarm
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie


class TestSwarm(unittest.TestCase):
    URI1 = 'uri1'
    URI2 = 'uri2'
    URI3 = 'uri3'

    def setUp(self):
        self.uris = [self.URI1, self.URI2, self.URI3]
        self.factory = MockFactory()

        self.sut = Swarm(self.uris, factory=self.factory)

    def test_that_instances_are_created(self):
        # Fixture

        # Test

        # Assert
        actual = len(self.factory.mocks)
        expected = len(self.uris)
        self.assertEqual(expected, actual)

    def test_that_all_links_are_opened(self):
        # Fixture

        # Test
        self.sut.open_links()

        # Assert
        for uri, mock in self.factory.mocks.items():
            mock.open_link.assert_called_once_with()

    def test_open_of_already_opened_swarm_raises_exception(self):
        # Fixture
        self.sut.open_links()

        # Test
        # Assert
        with self.assertRaises(Exception):
            self.sut.open_links()

    def test_failed_open_of_one_link_closes_all_and_raises_exception(self):
        # Fixture
        self.factory.mocks[self.URI2].open_link.side_effect = Exception()

        # Test
        # Assert
        with self.assertRaises(Exception):
            self.sut.open_links()

        for uri, mock in self.factory.mocks.items():
            mock.close_link.assert_called_once_with()

    def test_that_all_links_are_closed(self):
        # Fixture
        self.sut.open_links()

        # Test
        self.sut.close_links()

        # Assert
        for uri, mock in self.factory.mocks.items():
            mock.close_link.assert_called_once_with()

    def test_open_with_context_management(self):
        # Fixture

        # Test
        with Swarm(self.uris, factory=self.factory):
            pass

        # Assert
        for uri, mock in self.factory.mocks.items():
            mock.open_link.assert_called_once_with()
            mock.close_link.assert_called_once_with()

    def test_sequential_execution_without_arguments(self):
        # Fixture
        func = MagicMock()

        # Test
        self.sut.sequential(func)

        # Assert
        for uri, mock in self.factory.mocks.items():
            func.assert_any_call(mock)

    def test_sequential_execution(self):
        # Fixture
        func = MagicMock()
        args_dict = {
            self.URI1: ['cf1-arg1'],
            self.URI2: ['cf2-arg1'],
            self.URI3: ['cf3-arg1'],
        }

        cf1 = self.factory.mocks[self.URI1]
        cf2 = self.factory.mocks[self.URI2]
        cf3 = self.factory.mocks[self.URI3]

        # Test
        self.sut.sequential(func, args_dict=args_dict)

        # Assert
        func.assert_any_call(cf1, 'cf1-arg1')
        func.assert_any_call(cf2, 'cf2-arg1')
        func.assert_any_call(cf3, 'cf3-arg1')

    def test_parallel_execution_without_arguments(self):
        # Fixture
        func = MagicMock()

        # Test
        self.sut.parallel(func)

        # Assert
        for uri, mock in self.factory.mocks.items():
            func.assert_any_call(mock)

    def test_parallel_execution(self):
        # Fixture
        func = MagicMock()
        args_dict = {
            self.URI1: ['cf1-arg1'],
            self.URI2: ['cf2-arg1'],
            self.URI3: ['cf3-arg1'],
        }

        cf1 = self.factory.mocks[self.URI1]
        cf2 = self.factory.mocks[self.URI2]
        cf3 = self.factory.mocks[self.URI3]

        # Test
        self.sut.parallel(func, args_dict=args_dict)

        # Assert
        func.assert_any_call(cf1, 'cf1-arg1')
        func.assert_any_call(cf2, 'cf2-arg1')
        func.assert_any_call(cf3, 'cf3-arg1')

    def test_parallel_execution_with_exception(self):
        # Fixture
        func_fail = MagicMock()
        func_fail.side_effect = Exception()
        args_dict = {
            self.URI1: ['cf1-arg1'],
            self.URI2: ['cf2-arg1'],
            self.URI3: ['cf3-arg1'],
        }

        cf1 = self.factory.mocks[self.URI1]
        cf2 = self.factory.mocks[self.URI2]
        cf3 = self.factory.mocks[self.URI3]

        # Test
        self.sut.parallel(func_fail, args_dict=args_dict)

        # Assert
        func_fail.assert_any_call(cf1, 'cf1-arg1')
        func_fail.assert_any_call(cf2, 'cf2-arg1')
        func_fail.assert_any_call(cf3, 'cf3-arg1')

    def test_parallel_safe_execution_with_exception(self):
        # Fixture
        func_fail = MagicMock()
        func_fail.side_effect = Exception()
        args_dict = {
            self.URI1: ['cf1-arg1'],
            self.URI2: ['cf2-arg1'],
            self.URI3: ['cf3-arg1'],
        }

        # Test
        # Assert
        with self.assertRaises(Exception):
            self.sut.parallel_safe(func_fail, args_dict=args_dict)


class MockFactory:

    def __init__(self):
        self.mocks = {}

    def construct(self, uri):
        mock = MagicMock(spec=SyncCrazyflie, name='CF-' + uri)
        self.mocks[uri] = mock
        return mock


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from test.localization.lighthouse_fixtures import LighthouseFixtures
from test.localization.lighthouse_test_base import LighthouseTestBase

import numpy as np

from cflib.localization.ippe_cf import IppeCf
from cflib.localization.lighthouse_types import LhDeck4SensorPositions
from cflib.localization.lighthouse_types import Pose


class TestIppeCf(LighthouseTestBase):
    def setUp(self):
        self.fixtures = LighthouseFixtures()

    def test_that_pose_is_found(self):
        # Fixture
        pose_bs = Pose(t_vec=(0.0, 0.0, 0.0))
        pose_cf = Pose(t_vec=(1.0, 0.0, -1.0))

        U = LhDeck4SensorPositions.positions
        Q = self.fixtures.synthesize_angles(pose_cf, pose_bs).projection_pair_list()

        # The CF pose seen from the base station
        expected_0 = pose_cf

        # Not sure if (why) this is the expected mirror solution
        expected_1 = Pose.from_rot_vec(R_vec=(0.0, -np.pi / 2, 0.0), t_vec=pose_cf.translation)

        # Test
        actual = IppeCf.solve(U, Q)

        # Assert
        actual_pose_0 = Pose(actual[0].R, actual[0].t)
        self.assertPosesAlmostEqual(expected_0, actual_pose_0, places=3)

        actual_pose_1 = Pose(actual[1].R, actual[1].t)
        self.assertPosesAlmostEqual(expected_1, actual_pose_1, places=3)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import time
import unittest

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.mem import MemoryElement
from cflib.crazyflie.swarm import CachedCfFactory
from cflib.crazyflie.swarm import Swarm
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.crazyflie.syncLogger import SyncLogger
from sys_test.swarm_test_rig.rig_support import RigSupport


class TestMemoryMapping(unittest.TestCase):
    def setUp(self):
        cflib.crtp.init_drivers()
        self.test_rig_support = RigSupport()

    def test_memory_mapping_with_one_cf(self):
        # Fixture
        uri = self.test_rig_support.all_uris[0]
        self.test_rig_support.restart_devices([uri])
        cf = Crazyflie(rw_cache='./cache')

        # Test and Assert
        with SyncCrazyflie(uri, cf=cf) as scf:
            self.assert_memory_mapping(scf)

    def test_memory_mapping_with_all_cfs(self):
        # Fixture
        uris = self.test_rig_support.all_uris
        self.test_rig_support.restart_devices(uris)
        factory = CachedCfFactory(rw_cache='./cache')

        # Test and Assert
        with Swarm(uris, factory=factory) as swarm:
            swarm.parallel_safe(self.assert_memory_mapping)

    def test_memory_mapping_with_reuse_of_cf_object(self):
        # Fixture
        uri = self.test_rig_support.all_uris[0]
        self.test_rig_support.restart_devices([uri])
        cf = Crazyflie(rw_cache='./cache')

        # Test and Assert
        for connections in range(10):
            with SyncCrazyflie(uri, cf=cf) as scf:
                for mem_ops in range(5):
                    self.assert_memory_mapping(scf)

    # Utils

    def assert_memory_mapping(self, scf):
        mems = scf.cf.mem.get_mems(MemoryElement.TYPE_MEMORY_TESTER)
        count = len(mems)
        self.assertEqual(1, count, 'unexpected number of memories found')

        self.verify_reading_memory_data(mems)
        self.verify_writing_memory_data(mems, scf)

    def verify_writing_memory_data(self, mems, scf):
        self.wrote_data = False
        scf.cf.param.set_value('memTst.resetW', '1')
        time.sleep(0.1)
        mems[0].write_data(5, 1000, self._data_written)
        while not self.wrote_data:
            time.sleep(1)
        log_conf = LogConfig(name='memtester', period_in_ms=100)
        log_conf.add_variable('memTst.errCntW', 'uint32_t')
        with SyncLogger(scf, log_conf) as logger:
            for log_entry in logger:
                errorCount = log_entry[1]['memTst.errCntW']
                self.assertEqual(0, errorCount)
                break

    def verify_reading_memory_data(self, mems):
        self.got_data = False
        mems[0].read_data(5, 1000, self._data_read)
        while not self.got_data:
            time.sleep(1)

    def _data_read(self, mem):
        self.got_data = True

    def _data_written(self, mem, address):
        self.wrote_data = True
